import launchpad as lp
from launchpad.nodes import base
from launchpad import address as lp_address
from absl import logging

class ReverbHandle(base.Handle):

  def __init__(self, address: lp_address.Address, Client):
    self._address = address
    self.Client = Client

  def dereference(self):
    address = self._address.resolve()
    logging.info('Reverb client connecting to: %s', address)
    return self.Client(address)

class ReverbNode(lp.ReverbNode):
  """Inherits `ReverbNode` from launchpad and remove `worker_manager` from it."""



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
import os
import logging


class ScriptRunner:
    def __init__(self, cmd, cmd_args, pwd, log=None):
        self.cmd = cmd
        self.cmd_args = cmd_args
        self.pwd = pwd
        self.log = log if log else logging.getLogger()
        self.log.setLevel(logging.DEBUG)
        pass

    def run(self):
        self.log.info(f'Change path to {self.pwd}')
        os.chdir(self.pwd)
        cmd_str = self.cmd
        for arg in self.cmd_args:
            cmd_str += ' ' + str(arg)
        self.log.info(f'Run cmd: {cmd_str}')
        os.system(cmd_str)





import base64
from tlaunch.lp_k8s.util import map_opt
import logging
import string
from itertools import chain
from typing import Dict, List, Optional

import cloudpickle
from kubernetes import client, config
from launchpad.nodes.python.node import PyNode

from tlaunch.lp_k8s import addressing
from tlaunch.lp_k8s.config import Config
from tlaunch.lp_k8s.source import Source


def create_lp_job(node_groups: Dict[str, List[PyNode]],
                  job_name: Optional[str] = None,
                  sync_source: Optional[Source] = None,
                  namespace: Optional[str] = None,
                  kube_config: Optional[str] = None,
                  group_config: Optional[Dict[str, Config]] = None):
  job_name = validate_name(job_name)
  namespace = namespace or 'default'
  group_config = group_config or {}

  try:
    config.load_incluster_config()
  except Exception:
    config.load_kube_config(kube_config)

  for label, nodes in node_groups.items():
    for i, node in enumerate(nodes):
      # Run my own address binding strategy here.
      # Try to bind all rpc server to the same port on container with different dns domain. 
      addressing.bind_address_kubernetes(node.addresses, job_name, label, i)

  api = client.CustomObjectsApi()
  roles = {
      label: {
          'replicas':
          len(nodes),
          'template':
          client.V1PodTemplateSpec(
              metadata=client.V1ObjectMeta(labels={'app': job_name}),
              spec=group_config.get(label,
                                    Config(namespace=namespace)).set_job_name(job_name).build(),
          ),
          # REVIEW: sending exec over yaml or volume
          'executable':
          serialize(nodes)
      }
      for label, nodes in node_groups.items()
  }
  lp_job = {
      'apiVersion': 'tartrl.cn/v1alpha1',
      'kind': 'LpJob',
      'metadata': client.V1ObjectMeta(name=job_name),
      'spec': {
          'roles': roles,
          'source': map_opt(sync_source, lambda s: s.as_dict()) or None
      },
  }

  api.create_namespaced_custom_object(group='tartrl.cn',
                                      version='v1alpha1',
                                      namespace=namespace,
                                      plural='lpjobs',
                                      body=lp_job)
  logging.info(
      f'An lpjob named "{job_name}" has been created in namespace "{namespace}".'
  )


def serialize(nodes: List[PyNode]):
  data = cloudpickle.dumps([node.function for node in nodes])
  return base64.b64encode(data).decode()


def validate_name(name) -> str:
  if name is not None:
    name = name.lower().replace('_', '-')
    flag, new_name = is_valid_subdomain_filter(name)
    if not flag:
      logging.warn(
          f'"{name}" is not a valid subdomain name. '
          'A lowercase RFC 1123 subdomain must consist of lower case alphanumeric characters, '
          '\'-\' or \'.\', and must start and end with an alphanumeric character.'
      )
      logging.info(f'Stripping name to "{new_name}"')
      assert new_name is not None
      return new_name
    return name
  return name or 'launchpad'


def is_valid_subdomain_filter(domain):
  nums = map(str, range(0, 10))
  alphanumeric = set(chain(nums, string.ascii_lowercase))
  if domain[0] not in alphanumeric or domain[-1] not in alphanumeric:
    return False, None
  alphanumeric.update(['-', '.'])
  flag = all(ch in alphanumeric for ch in domain)
  new_domain = None
  if not flag:
    new_domain = str(filter(lambda ch: ch not in alphanumeric, domain))
  return flag, new_domain




class VolumeMount:
    def __init__(self, mount_path: str = None,
                 mount_propagation: str = None,
                 name: str = None,
                 read_only: bool = None,
                 sub_path: str = None,
                 sub_path_expr: str = None):

        self.mount_path = mount_path
        self.mount_propagation = mount_propagation
        self.name = name
        self.read_only = read_only
        self.sub_path = sub_path
        self.sub_path_expr = sub_path_expr



from typing import Dict, List, Optional, Tuple

from kubernetes import client

from tlaunch.lp_k8s.resource import Resource
from tlaunch.lp_k8s.volume_mount import VolumeMount
from tlaunch.lp_k8s.persistent_volume_claim_volume_source import PersistentVolumeClaimVolumeSource
from tlaunch.lp_k8s.util import map_opt

DEFAULT_PORTS = [('launchpad', 8001)]
DEFAULT_NAME = 'launchpad'
REVERB_IMAGE = 'docker.4pd.io/tlaunch/reverb:base'
DEFAULT_VOLUME_NAME = 'tdata'
DEFAULT_MNT_PATH = '/TData'
DEFAULT_COMMAND = ['bash', '-c', '/TData/code/setup.sh && python3 -u -mtlaunch.lp_k8s.process_entry']
DEBUG_COMMAND = ['bash', '-c', 'sleep 3600']
DEFAULT_PVC_NAME = 'jfsdemo'
DEFAULT_IMAGE_PULL_POLICY = 'IfNotPresent'


class Container:
    def __init__(self,
                 namespace: str,
                 image: Optional[str] = None,
                 command: Optional[List[str]] = None,
                 flags: List[str] = [],
                 resources: Optional[Resource] = None,
                 volume_mounts: Optional[List[VolumeMount]] = None,
                 env: Optional[Dict[str, int]] = None,
                 image_pull_policy: str = None,
                 ports: Optional[List[Tuple[str, int]]] = None):
        self.job_name = DEFAULT_NAME
        self.image = image or REVERB_IMAGE
        self.command = (command or DEFAULT_COMMAND) + flags
        self.resources = resources
        self.volume_mounts = volume_mounts or [
            VolumeMount(name=DEFAULT_VOLUME_NAME, mount_path=DEFAULT_MNT_PATH, sub_path='./' + namespace)]
        self.env = env
        self.image_pull_policy = image_pull_policy or DEFAULT_IMAGE_PULL_POLICY
        self.ports = DEFAULT_PORTS.extend(ports) if ports else DEFAULT_PORTS

    def build(self) -> client.V1Container:
        return client.V1Container(
            name=self.job_name,
            image=self.image,
            image_pull_policy=self.image_pull_policy,
            command=self.command,
            ports=map_opt(
                self.ports,
                lambda x: [client.V1ContainerPort(name=v[0],
                                                  container_port=v[1]) for v in x]
            ),
            resources=map_opt(
                self.resources,
                lambda x: client.V1ResourceRequirements(limits=x.limits,
                                                        requests=x.requests)) \
                if self.resources else None,
            volume_mounts=map_opt(
                self.volume_mounts,
                lambda t: [client.V1VolumeMount(mount_path=v.mount_path,
                                                mount_propagation=v.mount_propagation,
                                                name=v.name,
                                                read_only=v.read_only,
                                                sub_path=v.sub_path,
                                                sub_path_expr=v.sub_path_expr) for v in t]) \
                if self.volume_mounts else None,
            env=map_opt(
                self.env,
                lambda e: [client.V1EnvVar(name=k, value=v)
                           for k, v in e.items()]))

    def set_job_name(self, job_name: str) -> 'Container':
        self.job_name = job_name
        return self


class Volumer:
    def __init__(self,
                 namespace: str,
                 persistent_volume_claim: Optional[PersistentVolumeClaimVolumeSource] = None,
                 name: Optional[str] = None):
        claim_name = namespace + '-' + DEFAULT_PVC_NAME
        self.persistent_volume_claim = persistent_volume_claim or PersistentVolumeClaimVolumeSource(
            claim_name=claim_name)
        self.name = name or DEFAULT_VOLUME_NAME

    def build(self) -> client.V1Volume:
        return client.V1Volume(
            persistent_volume_claim=map_opt(
                self.persistent_volume_claim,
                lambda p: client.V1PersistentVolumeClaimVolumeSource(claim_name=p.claim_name,
                                                                     read_only=p.read_only)) \
                if self.persistent_volume_claim else None,
            name=self.name)


class Config:
    def __init__(self, namespace: str,
                 container: Optional[Container] = None,
                 volumer: Optional[Volumer] = None,
                 **kwargs):
        self.container = container or Container(namespace=namespace)
        self.volumer = volumer or Volumer(namespace=namespace)
        self.kwargs = kwargs

    def build(self) -> client.V1PodSpec:
        return client.V1PodSpec(**self.kwargs, containers=[self.container.build()],
                                volumes=[self.volumer.build()] if self.volumer else None)

    def set_job_name(self, job_name: str) -> 'Config':
        self.container.set_job_name(job_name)
        return self


class DefaultReverbConfig(Config):
    def __init__(self) -> None:
        self.container = Container(image=REVERB_IMAGE)



import socket
import os
import time
from typing import Callable, Iterable, Iterator, Optional, TypeVar


T = TypeVar('T')
U = TypeVar('U')


def map_opt(opt: Optional[T], fn: Callable[[T], U]) -> Optional[U]:
  return None if opt is None else fn(opt)

def get_hostname_by_lpjob(lpjob_name: str, node_name: str, id: int) -> str:
    return f'{lpjob_name}-{node_name}-{id}.{lpjob_name}'

def get_ipv4_by_hostname(hostname, log = None):
    ip_list =  list(i[4][0] for i in socket.getaddrinfo(hostname, 0)
        if i[0] is socket.AddressFamily.AF_INET and i[1] is socket.SocketKind.SOCK_RAW)

    if len(ip_list) == 0:
        if log:
            log.info('WARNING: {} not found'.format(hostname))
        return get_ipv4_by_hostname(hostname, log)

    return ip_list

def get_namespace() -> str:
    if 'TPOD_NAMESPACE' in os.environ:
        namespace = os.environ['TPOD_NAMESPACE']
    else:
        namespace = 'default'
    return namespace

def generate_tlaunchjob_name():
    return "tlaunchjob-"+str(time.strftime("%H:%M:%S", time.localtime()))



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
import os
from tlaunch.lp_k8s.runner.script_runner import ScriptRunner
from tlaunch.lp_k8s.util import get_ipv4_by_hostname
import logging


class TorchRunner(ScriptRunner):
    def __init__(self, rdzv_endpoint, cmd, cmd_args, pwd):
        log = logging.getLogger()
        log.setLevel(logging.DEBUG)
        rdzv_endpoint = get_ipv4_by_hostname(rdzv_endpoint, log)[0] + ':8002'
        cmd_args = [f'--rdzv_endpoint={rdzv_endpoint}'] + cmd_args
        super().__init__(cmd, cmd_args, pwd, log)
        pass






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import os
import logging

import subprocess

import atexit
import psutil

from launchpad.launch.worker_manager import ThreadWorker
from tlaunch.lp_ssh.launch.worker_manager import WorkerManager


class ServerProcess:
    def __init__(self, server, port, pid):
        self.server = server
        self.port = port
        self.pid = pid

    def is_alive(self):
        command = ['ssh','-p', self.port, self.server, 'python -c "import psutil;psutil.Process({})"'.format(self.pid)]
        try:
            subprocess.check_output(
                command,
                stderr=subprocess.STDOUT)

        except subprocess.CalledProcessError as e:
            logging.info('{}:{} is not alive'.format(self.server, self.pid))
            return False
        else:
            logging.info('{}:{} is alive'.format(self.server, self.pid))
            return True


class SSHWorkerManager(WorkerManager):
    def register_existing_ssh_process(self, name: str, server: str, port: str, pid: int):
        self._workers_count[name] += 1
        self._active_workers[name].append(ServerProcess(server, port=port,pid=pid))

    def _check_workers(self):
        """Checks status of running workers, terminate runtime in case of errors."""
        has_workers = False
        for label in self._active_workers:
            still_active = []
            for worker in self._active_workers[label]:
                active = True
                if isinstance(worker, ThreadWorker):
                    if not worker.thread.is_alive():
                        worker.thread.join()
                        if not self._stop_counter:
                            try:
                                worker.future.result()
                            except Exception as e:
                                if not self._first_failure and not self._stop_counter:
                                    self._first_failure = e
                        active = False
                elif isinstance(worker, subprocess.Popen):
                    try:
                        res = worker.wait(0)
                        active = False
                        if res and not self._first_failure and not self._stop_counter:
                            self._first_failure = RuntimeError('One of the workers failed.')
                    except subprocess.TimeoutExpired:
                        pass
                elif isinstance(worker, ServerProcess):
                    if not worker.is_alive():
                        active = False
                else:
                    try:
                        # We can't obtain return code of external process, so clean
                        # termination is assumed.
                        res = worker.wait(0)
                        active = False
                    except psutil.TimeoutExpired:
                        pass
                if active:
                    has_workers = True
                    still_active.append(worker)
            self._active_workers[label] = still_active
        if has_workers and self._first_failure and not self._stop_counter:
            self._stop()
        elif not has_workers:
            self._disable_alarm()


def launch_with_ssh_tmux_session(commands_to_launch,
                                 session_name_prefix=None):
    """Launch multiple CommandToLaunch tuples in a new ssh tmux session."""

    session_name_prefix = session_name_prefix or 'ssh_launch'

    return _launch_with_multiplex_ssh_session(commands_to_launch,
                                              session_name_prefix,
                                              'tmux')


def _launch_with_multiplex_ssh_session(commands_to_launch, session_name_prefix, multiplexer):
    # lp_ssh的主要操作在这里实现
    """Launch multiple CommandToLaunch tuples in a new multiplex session.

      Args:
        commands_to_launch: An iterable of `CommandToLaunch` namedtuples.
        session_name_prefix: Leading part of the name given to the new tmux session.
          If there is no existing session with this name, it will be used as-is,
          however if another session exists the name will be uniquified by appending
          an incrementing counter.
        multiplexer : tmux or byobu
    """
    # Make a new session with the unmodified name, if this fails add a suffix to
    # the name and retry.
    session_name_host_dict = {}
    ssh_execute = {}
    launch_dir = os.getcwd()

    for command_to_launch in commands_to_launch:
        session_name_host_dict[command_to_launch.host] = session_name_prefix
        ssh_execute[command_to_launch.host] = ['ssh', '-p', command_to_launch.port, command_to_launch.host]

    # 这一段主要就是用来在不同机器上启动tmux
    for host in session_name_host_dict:
        suffix_index = 0

        session_name = session_name_host_dict[host]

        while True:
            try:
                # 在远程机器上创建tmux
                exec_command_list = ssh_execute[host] + [multiplexer, 'new-session', '-d', '-s', session_name, '-c',
                                                         launch_dir]

                subprocess.check_output(exec_command_list,
                                        stderr=subprocess.STDOUT)
                # subprocess.check_output(
                #   ssh_execute[host] + [multiplexer, 'new-session', '-d', '-s', session_name],
                #   stderr=subprocess.STDOUT)

            except subprocess.CalledProcessError as e:
                if 'duplicate session' in e.output.decode():
                    logging.info('%r session %r already exists, trying to uniquify...',
                                 multiplexer, session_name)
                    session_name = '{}_{}'.format(session_name_prefix, suffix_index)
                    suffix_index += 1
                else:
                    raise e  # If `tmux new-session` failed for some other reason.
            else:
                # 输出tmux启动信息
                print('Launch tmux on {} with session name:{}'.format(ssh_execute[host][3],session_name) )
                break
        session_name_host_dict[host] = session_name


    for command_to_launch in commands_to_launch:
        # Apply command-specific overrides to environment variables.
        env_as_list = [
            f'{k}={v}' for k, v in command_to_launch.env_overrides.items()]

        # When the program is done, echo the command so it can be copy-pasted, and
        # then drop into a shell.

        command_str = subprocess.list2cmdline(env_as_list +
                                              command_to_launch.command_as_list)
        # inner_command = f'{command_str}; echo "{command_str}"; exec $SHELL; '
        inner_command = f'{command_str}; '

        window_name = command_to_launch.title

        session_name = session_name_host_dict[command_to_launch.host]
        # print("mamam", command_to_launch.host, ssh_execute[command_to_launch.host], session_name)
        # continue
        command = ssh_execute[command_to_launch.host] + [
            multiplexer,
            'new-window',
            '-t',
            session_name,
            '-n',
            window_name,
            '-c', launch_dir,
            inner_command,
        ]
        # Make the process block until it has completed.
        # print(command)
        print("{}\nHost {} session: {} window: {} execute {}".format('-'*20,ssh_execute[command_to_launch.host][3],session_name,window_name, inner_command[:-2]))
        subprocess.Popen(command)

    # print(
    #     f'Opened new {multiplexer} session called `{session_name}`. '
    #     f'If you are already in a tmux session, use `Ctrl+B W` as a '
    #     f'convenient way to switch to the new session. '
    #     f'Otherwise run \n\n  {multiplexer} a -t "{session_name}"\n\nTo change '
    #     f'the name of the tmux sessions use the `--tmux_session_name` flag. You '
    #     f'can terminate all the processes and the {multiplexer} session by '
    #     f'pressing Ctrl-C here.\n')

    def get_session_processes(host):
        # p = subprocess.run(ssh_execute[host] + [
        #   multiplexer, 'list-panes', '-t', session_name, '-s', '-F',
        #   '"#{pane_pid}"'], stdout=subprocess.PIPE, check=True)
        session_name = session_name_host_dict[host]
        p = subprocess.run(ssh_execute[host] + [
            multiplexer + ' list-panes -t ' + session_name + ' -s -F "#{pane_pid}"'], stdout=subprocess.PIPE,
                           check=True)

        return [int(pid) for pid in p.stdout.replace(b'"', b'').strip().split()]

    manager = SSHWorkerManager()
    atexit.register(manager.wait)

    for host in ssh_execute:
        for pid in get_session_processes(host):
            # continue
            manager.register_existing_ssh_process('ssh', host, ssh_execute[host][2], pid)

    # # kill session
    # kill_command = ssh_execute_str+" tmux kill-session -t "+ session_name
    # kill_command = kill_command.split(' ')
    # try:
    #     subprocess.check_output(kill_command,stderr=subprocess.STDOUT)
    # except subprocess.CalledProcessError as e:
    #     logging.info('kill {} failed!'.format(session_name))
    # else:
    #     logging.info('kill {} success!'.format(session_name))



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stops a Launchpad program."""

import os
import signal
from typing import Union

from absl import logging

from tlaunch.lp_ssh import context
import subprocess

def make_program_stopper(launch_type: Union[str, context.LaunchType]):
  """Returns a callable that stops the Launchpad program.

  Args:
    launch_type: launch_type with which the program stopper is used.

  Returns:
    A callable. When called, it stops the running program.
  """
  launch_type = context.LaunchType(launch_type)

  if launch_type in [
      context.LaunchType.LOCAL_MULTI_PROCESSING,
      context.LaunchType.LOCAL_MULTI_THREADING,
      context.LaunchType.TEST_MULTI_PROCESSING,
      context.LaunchType.TEST_MULTI_THREADING
  ]:
    launcher_process_id = os.getpid()

    def ask_launcher_for_termination(mark_as_completed=False):
      del mark_as_completed
      try:
        os.kill(launcher_process_id, signal.SIGTERM)
      except:
        print('can\'t kill {}'.format(launcher_process_id))

    return ask_launcher_for_termination

  if launch_type == context.LaunchType.SSH_MULTI_PROCESSING:
    def ask_launcher_for_termination(mark_as_completed=False):
      del mark_as_completed
      # kill session
      show_name_commad = "tmux display-message -p '#S'"
      show_name_commad = show_name_commad.split(' ')
      session_name = subprocess.check_output(show_name_commad, stderr=subprocess.STDOUT).decode(
        encoding='utf-8').strip().strip('\'')

      kill_command = "tmux kill-session -t " + session_name
      kill_command = kill_command.split(' ')
      print(kill_command)
      try:
        subprocess.check_output(kill_command, stderr=subprocess.STDOUT)
      except subprocess.CalledProcessError as e:
        logging.info('kill {} failed!'.format(session_name))
      else:
        logging.info('kill {} success!'.format(session_name))

    return ask_launcher_for_termination

  raise NotImplementedError(f'{launch_type} is not yet supported.')



from typing import Any, Dict, List, Optional, Sequence, Union, cast

import launchpad as lp
import os
from launchpad.nodes.python.node import PyNode

from tlaunch.lp_k8s import client
from tlaunch.lp_k8s.config import Config
from tlaunch.lp_k8s.source import Source


def launch(programs: Union[lp.Program, Sequence[lp.Program]],
           source: Optional[Source] = None,
           namespace: Optional[str] = None,
           kube_config: Optional[str] = None,
           group_config: Optional[Dict[str, Config]] = None) -> Any:
    # Pretend to be launching with local multiprocess here.
    launch_type = lp.LaunchType.LOCAL_MULTI_PROCESSING

    if not isinstance(programs, Sequence):
        programs = cast(Sequence[lp.Program], [programs])

    if namespace is None:
        if 'TPOD_NAMESPACE' in os.environ:
            namespace = os.environ['TPOD_NAMESPACE']
        else:
            namespace = 'default'

    # NOTE: original launchpad is launching only one single program here
    program = programs[0]

    launch_kubernetes(program, launch_type, source, namespace, kube_config,
                      group_config)


def launch_kubernetes(
        program: lp.Program,
        launch_type: lp.LaunchType = lp.LaunchType.LOCAL_MULTI_PROCESSING,
        source: Optional[Source] = None,
        namespace: Optional[str] = None,
        kube_config: Optional[str] = None,
        group_config: Optional[Dict[str, Config]] = None):
    # Namespace and name of Lpjob should be stored in launch_config
    l_config = {
        "kube_namespace": namespace or "default",
        "lpjob_name": client.validate_name(program.name),
    }

    for label, nodes in program.groups.items():
        for node in nodes:
            node._initialize_context(launch_type, launch_config=l_config)

    for label, nodes in program.groups.items():
        for node in nodes:
            for handle in node._input_handles:
                handle.connect(node, label)

    groups = cast(Dict[str, List[PyNode]], program.groups)

    client.create_lp_job(groups, program.name, source, namespace, kube_config,
                         group_config)



"""
Entry of a PythonNode worker.
This is mostly copied from launchpad, but slimmer.
Note that this is only usable with my implementation of lp-operator.
"""

import builtins
import contextlib
import logging
import os
import sys
import platform

from types import ModuleType

import cloudpickle


class DummyModule(ModuleType):
  def __getattr__(self, key):
    return None

  __all__ = []


# def tryimport(name, globals={}, locals={}, fromlist=[], level=0):
#   try:
#     return realimport(name, globals, locals, fromlist, level)
#   except ImportError:
#     print(f'Possibly missing import "{name}"')
#     return DummyModule(name)
#
#
# realimport, builtins.__import__ = builtins.__import__, tryimport

CONFIG_DIR = '/TData/cache'
SOURCE_DIR = '/workdir'
CODE_DIR = '/TData/code'

def main():
  # Allow for importing modules from the current directory.
  node_name, _, lp_task_id = platform.node().rpartition('-')
  job_name, _, node_name = node_name.rpartition('-')
  CACHE_DIR = os.path.join(CONFIG_DIR, job_name)
  task_id = int(lp_task_id)
  # os.system(f'bash {CODE_DIR}/setup.sh')
  if os.path.exists(SOURCE_DIR):
    for dirpath in os.listdir(SOURCE_DIR):
      path = os.path.join(SOURCE_DIR, dirpath)
      if os.path.islink(path):
        sys.path.append(path)

  functions = cloudpickle.load(open(os.path.join(CACHE_DIR, node_name), 'rb'))

  with contextlib.suppress():  # no-op context manager
    functions[task_id]()


if __name__ == '__main__':
  main()



class PersistentVolumeClaimVolumeSource:
    def __init__(self, claim_name: str = None,
                 read_only: bool = None):
        self.claim_name = claim_name
        self.read_only = read_only



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
from typing import Dict

import numpy as np
import torch
import reverb

from dill import dumps, loads

class Client(reverb.Client):
    def insert_weight(self, model_weight, priorities: Dict[str, float]):
        model_values = []
        episode = -1
        for model_key in model_weight:
            if model_key == 'episode':
                episode = model_weight['episode']
                continue

            keys = model_weight[model_key].keys()
            for key in keys:
                model_values.append(model_weight[model_key][key].cpu())

        model_values.append(episode)

        self.insert(data=model_values, priorities=priorities)


    # def sample_weight(self,model_keys,target_episode, table: str):
    #     weight_episode = target_episode - 1
    #     while weight_episode < target_episode:
    #         new_values = list(self.sample(table=table, num_samples=1))[0][0].data
    #         weight_episode = new_values[-1]
    #
    #     new_model_weight = {}
    #     for i, key in enumerate(model_keys):
    #         current_keys = key.split('@')
    #         model_key, sub_key = current_keys[0], current_keys[1]
    #         if model_key in new_model_weight:
    #             new_model_weight[model_key][sub_key] = torch.tensor(new_values[i])
    #         else:
    #             new_model_weight[model_key] = {sub_key: torch.tensor(new_values[i])}
    #     return new_model_weight, weight_episode


    def send_info(self,info):
      data = dumps(info)
      self.insert(data=data, priorities={'info':1})

    def get_info(self):
      feteched_data = np.array(list(self.sample(table='info', num_samples=1))[0][0].data[0], dtype=bytes)
      return loads(feteched_data)



from typing import List

import launchpad as lp

from tlaunch.lp_k8s.address import KubernetesAddressBuilder


def bind_address_kubernetes(addresses: List[lp.Address], lp_job_name: str,
                            group_name: str, id: int):
  """Bind addresses for kubernetes launch."""

  for address in addresses:
    address.bind(KubernetesAddressBuilder(lp_job_name, group_name, id))



import os
import sys
import logging

from typing import Any, Dict, Optional, Sequence, Union, cast,Text
from absl import flags

from launchpad import program as lp_program
from launchpad.launch.run_locally.run_locally import _LOCAL_LAUNCHER_MAP, SEPARATE_TERMINAL_MODES, TERMINALS_FOR_X
from launchpad.launch.run_locally.run_locally import SEPARATE_TERMINAL_XTERM,SEPARATE_TERMINAL_GNOME_TERMINAL_WINDOWS
from launchpad.launch.run_locally.run_locally import SEPARATE_TERMINAL_TMUX_SESSION,SEPARATE_TERMINAL_CURRENT_TERMINAL

from launchpad.launch.run_locally import feature_testing

from tlaunch.lp_ssh import context
from .run_ssh.launch_ssh_tmux import launch_with_ssh_tmux_session

FLAGS = flags.FLAGS

SEPARATE_SSH_TERMINAL_TMUX_SESSION = 'ssh_tmux_session'
_LOCAL_LAUNCHER_MAP.update({SEPARATE_SSH_TERMINAL_TMUX_SESSION:launch_with_ssh_tmux_session})
SEPARATE_TERMINAL_MODES = SEPARATE_TERMINAL_MODES+(SEPARATE_SSH_TERMINAL_TMUX_SESSION,)

def _get_terminal(given_terminal: Optional[Text]):
  """Returns the terminal for local launch based on X & command availability.

  By order of priority it will:
  - use the provided `given_terminal`
  - default to the shell environment variable `LAUNCHPAD_LAUNCH_LOCAL_TERMINAL`
    if set
  - or select the first supported option in: Gnome, Tmux, Xterm and current
    terminal.

  Args:
    given_terminal: The terminal identifier to use or `None`.

  Returns:
    One of the legal terminal modes (a string in SEPARATE_TERMINAL_MODES) based
    on the priority described above.
  """

  if (given_terminal is not None and
      given_terminal not in SEPARATE_TERMINAL_MODES):
    raise ValueError('`terminal` got a mode that it does not '
                     'understand %r. Please choose from %r.' %
                     (given_terminal, SEPARATE_TERMINAL_MODES))
  terminal = given_terminal or os.environ.get('LAUNCHPAD_LAUNCH_LOCAL_TERMINAL',
                                              None)
  # Set terminal to None, if the chosen terminal cannot be used because we are
  # running without X.
  if not feature_testing.has_x() and terminal in TERMINALS_FOR_X:
    logging.info('Not using %s to launch, since DISPLAY is not set.', terminal)
    terminal = None

  if terminal is None:
    if feature_testing.has_gnome_terminal():
      terminal = SEPARATE_TERMINAL_GNOME_TERMINAL_WINDOWS
    elif feature_testing.has_tmux():
      terminal = SEPARATE_TERMINAL_TMUX_SESSION
    elif feature_testing.has_xterm():
      terminal = SEPARATE_TERMINAL_XTERM

    # Examine the type of terminal and explain why it is chosen.
    if terminal is None:
      logging.info('Launching in the same console since we cannot find '
                   'gnome-terminal, tmux, or xterm.')
      terminal = SEPARATE_TERMINAL_CURRENT_TERMINAL
    else:
      logging.info(
          'Launching with %s because the `terminal` launch option '
          'is not explicitly specified. To remember your preference '
          '(assuming tmux_session is the preferred option), either: \n'
          '1. Pass the `terminal` launch option (e.g., '
          '`lp.launch(program, terminal="tmux_session")`).\n'
          '2. Set the following in your bashrc to remember your '
          'preference:\n'
          '    export LAUNCHPAD_LAUNCH_LOCAL_TERMINAL=tmux_session', terminal)

  return terminal

def launch(
    programs: Union[lp_program.Program, Sequence[lp_program.Program],],
    local_resources: Optional[Dict[str, Any]] = None,
    terminal: str = 'gnome-terminal',session_name_prefix=None
) -> Any:

  if not FLAGS.is_parsed():
    FLAGS(sys.argv, known_only=True)


  if not isinstance(programs, Sequence):
    programs = cast(Sequence[lp_program.Program], [programs])

  program = programs[0]

  # return launch_local_multiprocessed.launch(program, local_resources, terminal)

  local_resources = local_resources or {}
  for label, nodes in program.groups.items():

    launch_config = local_resources.get(label, None)

    for node in nodes:
      if terminal in ['tmux_session','current_terminal']:
        node._initialize_context(
          context.LaunchType.LOCAL_MULTI_PROCESSING,
          launch_config=launch_config)
      elif terminal == 'ssh_tmux_session':
        node._initialize_context(
          context.LaunchType.SSH_MULTI_PROCESSING,
          launch_config=launch_config)

  # Notify the input handles
  for label, nodes in program.groups.items():
    for node in nodes:
      for handle in node._input_handles:
        handle.connect(node, label)

  # Bind addresses
  for node in program.get_all_nodes():
    node.bind_addresses()


  commands = []

  for label, nodes in program.groups.items():
    # to_executables() is a static method, so we can call it from any of the
    # nodes in this group.

    # pytype: disable=wrong-arg-count
    commands.extend(nodes[0].to_executables(nodes, label,
                                            nodes[0]._launch_context))


    # pytype: enable=wrong-arg-count

  for command in commands:
    print("{}:{} to use python: {}".format(command.host,command.port, command.command_as_list[0]) )
    if not os.access(command.command_as_list[0], os.X_OK):
      raise ValueError("Unable to execute '%s'" % command.command_as_list[0])

  # terminal_name是个字符串，代表terminal的名字
  terminal_name = _get_terminal(terminal)

  # 通过_LOCAL_LAUNCHER_MAP字典来获取需要执行的函数
  # 如果是terminal_name == 'ssh_tmux_session'，则使用 launch_with_ssh_tmux_session

  return _LOCAL_LAUNCHER_MAP[terminal_name](commands,session_name_prefix)






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import os
import json
from pathlib import Path

from typing import Any, List, Mapping, Optional, Union, Sequence, Tuple
import shutil

import atexit

import cloudpickle
import tempfile

from launchpad.nodes.python.local_multi_processing import PythonProcess, _to_cmd_arg, flags_utils
from tlaunch.lp_ssh import context

from tlaunch.lp_ssh.launch.ssh_multi_processing import commands as mp_commands

from .python.node import PyClassNode

_DATA_FILE_NAME = 'job.pkl'


def to_ssh_multiprocessing_executables(
        nodes: Sequence[Any], label: str, launch_config: PythonProcess,
        pdb_post_mortem: bool) -> List[mp_commands.Command]:
    """Returns a list of `Command`s objects for the given `PyNode`s."""

    launch_config = launch_config or PythonProcess()
    if not isinstance(launch_config, PythonProcess):
        raise ValueError(
            'Launch config for {} must be a PythonProcess.'.format(label))

    #
    # mannully set entry script path

    # 在分布式系统中，这个data_file有两种参考的实现形式，一种是通过共享文件的形式，所有机器能访问一个统一的文件路径
    # 第二个实现方式是通过打包文件，然后把整个文件发送到远程机器上，远程机器加在本地文件即可
    # 这里我们采用第一种，通过共享文件夹路径的形式来加载'process_entry.py'和'job.pkl'

    if nodes[0].share_entry_script_path:
        entry_script_path = nodes[0].share_entry_script_path
    else:
        entry_script_path = os.path.join(os.path.dirname(__file__), 'python',
                                         'process_entry.py')
    if nodes[0].share_temp_dir:
        if not Path(nodes[0].share_temp_dir).exists():
            os.makedirs(nodes[0].share_temp_dir)
        tmp_dir = tempfile.mkdtemp(dir=nodes[0].share_temp_dir)
    else:
        tmp_dir = tempfile.mkdtemp()

    # debug = True
    # if debug:
    #   # 这里为tlaunch分布式执行时，代码的入口，后面需要改成自动加载这个入口
    #   entry_script_path = '/home/shiyu/mfs/Task/TARTRL/github/tmarl/tmarl/tlaunch/lp_ssh/nodes/python/process_entry.py'
    #   tmp_dir = tempfile.mkdtemp(dir=os.path.join(os.getenv("HOME"), 'mfs/tmp/'))

    rm_tmp_dir = False
    if rm_tmp_dir:
        atexit.register(shutil.rmtree, tmp_dir, ignore_errors=True)


    data_file_path = os.path.join(tmp_dir, _DATA_FILE_NAME)
    with open(data_file_path, 'wb') as f:
        cloudpickle.dump([node.function for node in nodes], f,protocol=4)

    commands = []
    if hasattr(nodes[0], "host"):
        group_host = nodes[0].host
        group_port = nodes[0].port
    else:
        group_host = "localhost"
        group_port = "22"

    for task_id, node in enumerate(nodes):
        assert not hasattr(node, "host") or group_host == node.host, 'Nodes in the same group must have the same host. Node {} in group {} should run on host {} instead of {}'.format(
            task_id, label, group_host, node.host)
        command_as_list = [
            launch_config.absolute_interpreter_path, entry_script_path
        ]

        # Arguments to pass to the script
        for key, value in launch_config.args.items():
            command_as_list.extend(_to_cmd_arg(key, value))

        # Find flags and pre-populate their definitions, as these definitions are
        # not yet ready in the entry script.
        flags_to_populate = flags_utils.get_flags_to_populate(
            list(launch_config.args.items()))
        if flags_to_populate:
            command_as_list.extend(
                _to_cmd_arg('flags_to_populate', json.dumps(flags_to_populate)))

        command_as_list.extend([
            '--data_file', data_file_path,
            '--lp_task_id', str(task_id),
        ])
        command = mp_commands.Command(command_as_list, launch_config.env,
                                      label + '/' + str(task_id), group_host, group_port)

        commands.append(command)

    return commands


class SSHNode(PyClassNode):
    share_entry_script_path: str = None
    share_temp_dir: str = None

    def run(self) -> None:
        super().run()

    def to_host(self, host):
        if ":" in host:
            self.host = host.split(":")[0]
            self.port = host.split(":")[1]
        else:
            self.host = host
            self.port = "22"
        return self

    @staticmethod
    def to_executables(nodes, label, launch_context):
        return to_ssh_multiprocessing_executables(
            nodes, label, launch_context.launch_config, pdb_post_mortem=False)



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Addressing for PyNodes."""

import getpass
import os
import typing
from typing import Any, List, Optional

from absl import flags
from absl import logging
from tlaunch.lp_ssh import address as lp_address


FLAGS = flags.FLAGS

def bind_addresses_local(addresses: List[lp_address.Address]):
  """Binds addresses for the local launch."""

  for address in addresses:
    address.bind(lp_address.LocalAddressBuilder())

def bind_addresses_host(addresses: List[lp_address.Address]):
  """Binds addresses for the local launch."""

  for address in addresses:
    address.bind(lp_address.HostAddressBuilder(address.host,address.port))



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Entry of a PythonNode worker."""


import contextlib
import json
import os
import sys

import traceback

from absl import app
from absl import flags
from absl import logging
import cloudpickle
from tlaunch.lp_ssh.launch import worker_manager
import six


FLAGS = flags.FLAGS

flags.DEFINE_integer(
    'lp_task_id', None, 'a list index deciding which '
    'worker to run. given a list of workers (obtained from the'
    ' data_file)')
flags.DEFINE_string('data_file', '',
                    'Pickle file location with entry points for all nodes')
flags.DEFINE_string(
    'lp_job_name', '',
    'The name of the job, used to access the correct pickle file resource when '
    'using the new launch API')
flags.DEFINE_string(
    'init_file', '', 'Pickle file location containing initialization module '
    'executed for each node prior to an entry point')
flags.DEFINE_string('flags_to_populate', '{}', '')

_FLAG_TYPE_MAPPING = {
    str: flags.DEFINE_string,
    six.text_type: flags.DEFINE_string,
    float: flags.DEFINE_float,
    int: flags.DEFINE_integer,
    bool: flags.DEFINE_boolean,
    list: flags.DEFINE_list,
}


def _populate_flags():
  """Populate flags that cannot be passed directly to this script."""
  FLAGS(sys.argv, known_only=True)

  flags_to_populate = json.loads(FLAGS.flags_to_populate)
  for name, value in flags_to_populate.items():
    value_type = type(value)
    if value_type in _FLAG_TYPE_MAPPING:
      flag_ctr = _FLAG_TYPE_MAPPING[value_type]
      logging.info('Defining flag %s with default value %s', name, value)
      flag_ctr(
          name,
          value,
          'This flag has been auto-generated.',
          allow_override=True)

  # JAX doesn't use absl flags and so we need to forward absl flags to JAX
  # explicitly. Here's a heuristic to detect JAX flags and forward them.
  for arg in sys.argv:
    if arg.startswith('--jax_'):
      try:
        # pytype:disable=import-error
        import jax
        # pytype:enable=import-error
        jax.config.parse_flags_with_absl()
        break
      except ImportError:
        pass


def _get_task_id():
  """Returns current task's id."""
  if FLAGS.lp_task_id is None:
    # Running under Vertex AI...
    cluster_spec = os.environ.get('CLUSTER_SPEC', None)
    return json.loads(cluster_spec).get('task').get('index')

  return FLAGS.lp_task_id


def main(_):
  # Allow for importing modules from the current directory.
  sys.path.append(os.getcwd())
  data_file = FLAGS.data_file
  init_file = FLAGS.init_file

  if os.environ.get('TF_CONFIG', None):
    # For GCP runtime log to STDOUT so that logs are not reported as errors.
    logging.get_absl_handler().python_handler.stream = sys.stdout

  if init_file:
    init_function = cloudpickle.load(open(init_file, 'rb'))
    init_function()
  f = open(data_file, 'rb')
  functions = cloudpickle.load(f)
  task_id = _get_task_id()

  # Worker manager is used here to handle termination signals and provide
  # preemption support.
  worker_manager.WorkerManager(
      register_in_thread=True)

  with contextlib.suppress():  # no-op context manager
    try:
      functions[task_id]()
  
    except Exception as e:
          traceback.print_exc()
          # 或者得到堆栈字符串信息
          # info = traceback.format_exc()
          # s = sys.exc_info()
          # print("Error '%s' happened on line %d" % (s[1], s[2].tb_lineno))
          answer = input("Some thing is wrong, stop all[yes or no]?:")
          if answer == 'yes':
              pass

import torch

if __name__ == '__main__':
  torch.multiprocessing.set_start_method('spawn')     # 多进程中使用cuda
  _populate_flags()
  app.run(main)



from typing import Dict, Optional

import os
import tarfile
import tempfile
from minio import Minio


class Source:
  def source_type(self) -> str:
    return 'unknown'

  def fields(self) -> Dict[str, Optional[str]]:
    return {}

  def as_dict(self) -> Dict[str, Dict[str, Optional[str]]]:
    return {self.source_type(): self.fields()}


class GitSource(Source):
  def __init__(self, url: str, username: Optional[str],
               password: Optional[str]) -> None:
    self.url = url
    self.username = username
    self.password = password

  def source_type(self) -> str:
    return 'git'

  def fields(self) -> Dict[str, Optional[str]]:
    return {
        'url': self.url,
        'username': self.username,
        'password': self.password
    }


class MinioSource(Source):
  def __init__(self,
               endpoint: str,
               bucket: str,
               path: str,
               access_key: Optional[str] = None,
               secret_key: Optional[str] = None,
               upload_local_path: Optional[str] = None) -> None:
    self.endpoint = endpoint
    self.bucket = bucket
    self.path = path
    self.access_key = access_key
    self.secret_key = secret_key

    if upload_local_path:
      # this will overwrite existing object
      dirpath = tempfile.mkdtemp()
      tarball = os.path.join(dirpath, 'source.tar.gz')
      with tarfile.open(tarball, mode='w:gz') as archive:
        archive.add(upload_local_path)
      # secure=False here to avoid `ssl wrong version number` error
      client = Minio(endpoint=endpoint,
                     access_key=access_key,
                     secret_key=secret_key,
                     secure=False)
      client.fput_object(bucket_name=bucket,
                         object_name=path,
                         file_path=tarball)

  def source_type(self) -> str:
    return 'minio'

  def fields(self) -> Dict[str, Optional[str]]:
    return {
        'endpoint': self.endpoint,
        'bucket': self.bucket,
        'path': self.path,
        'accessKey': self.access_key,
        'secretKey': self.secret_key,
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2022 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import subprocess
def kill_tmux_on_host(host_origin,session_name_prefix,kill_all):
    multiplexer = "tmux"

    if ":" in host_origin:
        host = host_origin.split(":")[0]
        port = host_origin.split(":")[1]
    else:
        host = host_origin
        port  = "22"

    if kill_all:
        try:
            print("kill all sessions on {}:{}".format(host, port))
            if host == "localhost":
                exec_command_list = [multiplexer, 'kill-server']
            else:
                exec_command_list = ['ssh', '-p', port, host] + [multiplexer, 'kill-server']
            subprocess.check_output(exec_command_list)

        except:
            pass
    else:
        try:

            if host == "localhost":

                exec_command_list = [multiplexer, 'ls']

            else:
                exec_command_list = ['ssh', '-p', port, host] + [multiplexer, 'ls']

            output = subprocess.check_output(exec_command_list,
                                    stderr=subprocess.STDOUT)

            session_names = []
            for line in output.decode().split('\n'):

                if ":" in line:
                    session_names.append(line.split(":")[0])
            # print(session_names)

            for name in session_names:
                if name.startswith(session_name_prefix):
                    print("kill {} on {}:{}".format(name, host,port))
                    if host == "localhost":
                        exec_command_list = [multiplexer, 'kill-session', '-t', name]
                    else:
                        exec_command_list = ['ssh', '-p', port, host] + [multiplexer, 'kill-session', '-t', name]
                    print("\t"+" ".join(exec_command_list))
                    subprocess.check_output(exec_command_list)
        except:
            pass

def kill_tmux(host_list, session_name_prefix,kill_all=False):
    for host in host_list:
        kill_tmux_on_host(host,session_name_prefix,kill_all)




import base64

import logging
import string
from itertools import chain
from typing import Dict, List, Optional

import cloudpickle

from launchpad.nodes.python.node import PyNode
from launchpad.launch.run_locally import run_locally



def create_ssh_job(program,terminal: str = 'gnome-terminal'):
  commands = []
  for label, nodes in program.groups.items():
    # to_executables() is a static method, so we can call it from any of the
    # nodes in this group.
    # pytype: disable=wrong-arg-count
    commands.extend(nodes[0].to_executables(nodes, label,
                                            nodes[0].launch_context))
    # pytype: enable=wrong-arg-count
  return run_locally.run_commands_locally(commands, terminal)




def serialize(nodes: List[PyNode]):
  data = cloudpickle.dumps([node.function for node in nodes])
  return base64.b64encode(data).decode()


def validate_name(name) -> str:
  if name is not None:
    name = name.lower().replace('_', '-')
    flag, new_name = is_valid_subdomain_filter(name)
    if not flag:
      logging.warn(
          f'"{name}" is not a valid subdomain name. '
          'A lowercase RFC 1123 subdomain must consist of lower case alphanumeric characters, '
          '\'-\' or \'.\', and must start and end with an alphanumeric character.'
      )
      logging.info(f'Stripping name to "{new_name}"')
      assert new_name is not None
      return new_name
    return name
  return name or 'launchpad'


def is_valid_subdomain_filter(domain):
  nums = map(str, range(0, 10))
  alphanumeric = set(chain(nums, string.ascii_lowercase))
  if domain[0] not in alphanumeric or domain[-1] not in alphanumeric:
    return False, None
  alphanumeric.update(['-', '.'])
  flag = all(ch in alphanumeric for ch in domain)
  new_domain = None
  if not flag:
    new_domain = str(filter(lambda ch: ch not in alphanumeric, domain))
  return flag, new_domain



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stop a running Launchpad program."""

from absl import flags


from tlaunch.lp_ssh.launch import worker_manager

from tlaunch.lp_ssh import context

FLAGS = flags.FLAGS


def stop():
  """Terminates the entire experiment."""
  context.get_context().program_stopper()


def wait_for_stop():
  """Waits for program's termination/preemption signal."""
  worker_manager.get_worker_manager().wait_for_stop()



# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""WorkerManager handles thread and process-based runtimes."""

import atexit
import collections
from concurrent import futures
import os
import signal
import subprocess
import sys
import threading
import time
from typing import Optional, Sequence, Text

from absl import flags
from absl import logging
from absl.testing import absltest
from tlaunch.lp_ssh.flags import flags as lp_flags
import psutil
import termcolor

FLAGS = flags.FLAGS

ThreadWorker = collections.namedtuple('ThreadWorker', ['thread', 'future'])

_WORKER_MANAGERS = threading.local()
_HAS_MAIN_MANAGER = False


def get_worker_manager():
  manager = getattr(_WORKER_MANAGERS, 'manager', None)
  if not manager:
    raise RuntimeError('Worker manager is only available from the Launchpad''s '
                       'program node thread.')
  return manager


def register_signal_handler(sig, handler):
  """Registers a signal handler."""
  return signal.signal(sig, handler)


def remove_signal_handler(sig, handler):
  """Unregisters a signal handler."""
  if sys.is_finalizing():
    # See https://bugs.python.org/issue26133.
    return

  return signal.signal(sig, handler)


def wait_for_stop(timeout_secs: Optional[float] = None):
  """Blocks until termination of the node's program starts or timeout passes.

  Args:
    timeout_secs: Floating point number specifying a timeout for the operation,
        in seconds. If not provided, timeout is infinite.

  Returns:
    True if program is being terminated, False if timeout was reached.

  Usage examples:
  - Perform cleanup at the end of the run:
    start_server()
    lp.wait_for_stop()
    stop_server()
    checkpoint()

  - Perform some work until program is terminated:
    while not lp.wait_for_stop(0): # Return immediately.
      ... do some work ...

  - Perform some task every 5 seconds:
    while not lp.wait_for_stop(5.0):
      ... perform periodic task ...
  """
  return get_worker_manager().wait_for_stop(timeout_secs)


def stop_event():
  """Returns a threading.Event used to wait for termination signal on a Program.

  Usage examples:
  - Perform cleanup at the end of the run:
    start_server()
    lp.stop_event().wait()
    stop_server()
    checkpoint()
  """
  return get_worker_manager().stop_event()

class WorkerManager:
  """Encapsulates running threads and processes of a Launchpad Program."""

  def __init__(
      self,
      kill_main_thread=True,
      register_in_thread=False,
      register_signals=True):
    """Initializes a WorkerManager.

    Args:
      kill_main_thread: When set to false try not to kill the launcher while
        killing workers. This is not possible when thread workers run in the
        same process.
      register_in_thread: TODO
      register_signals: Whether or not to register signal handlers.
    """
    self._mutex = threading.Lock()
    self._termination_notice_secs = -1
    handle_user_stop = False
    global _HAS_MAIN_MANAGER
    # Make the first created worker manager the main manager, which handles
    # signals.
    if not _HAS_MAIN_MANAGER:
      self._termination_notice_secs = FLAGS.lp_termination_notice_secs
      handle_user_stop = True
      _HAS_MAIN_MANAGER = True
    self._active_workers = collections.defaultdict(list)
    self._workers_count = collections.defaultdict(lambda: 0)
    self._first_failure = None
    self._stop_counter = 0
    self._alarm_enabled = False
    self._kill_main_thread = kill_main_thread
    self._stop_event = threading.Event()
    self._main_thread = threading.current_thread().ident
    self._sigterm_handler = None
    self._sigquit_handler = None
    self._sigalrm_handler = None
    if register_signals:
      self._sigterm_handler = register_signal_handler(signal.SIGTERM,
                                                      self._sigterm)
      self._sigquit_handler = register_signal_handler(signal.SIGQUIT,
                                                      self._sigquit)
    if handle_user_stop:
      register_signal_handler(
          signal.SIGINT, lambda sig=None, frame=None: self._stop_by_user())
    if register_in_thread:
      _WORKER_MANAGERS.manager = self

  def _disable_signals(self):
    if not hasattr(self,'_sigterm_handler'):
      self._sigterm_handler = None
    if not hasattr(self, '_sigquit_handler'):
      self._sigquit_handler = None

    self._disable_alarm()
    if self._sigterm_handler is not None:
      remove_signal_handler(signal.SIGTERM, self._sigterm_handler)
      self._sigterm_handler = None
    if self._sigquit_handler is not None:
      remove_signal_handler(signal.SIGQUIT, self._sigquit_handler)
      self._sigquit_handler = None

  def _sigterm(self, sig=None, frame=None):
    """Handles SIGTERM by stopping the workers."""
    if callable(self._sigterm_handler):
      self._sigterm_handler(sig, frame)
    self._stop()

  def _sigquit(self, sig=None, frame=None):
    if callable(self._sigquit_handler):
      self._sigquit_handler(sig, frame)
    self._kill()

  def wait_for_stop(self, timeout_secs: Optional[float] = None):
    """Blocks until managed runtime is terminating or timeout is reached."""
    return self._stop_event.wait(timeout_secs)

  def stop_event(self):
    """Returns an event used to wait for termination signal on a Program."""
    return self._stop_event

  def thread_worker(self, name, function):
    """Registers and starts a new thread worker.

    Args:
      name: Name of the worker group.
      function: Entrypoint function to execute in a worker.
    """
    with self._mutex:
      future = futures.Future()

      def run_inner(f=function, future=future, manager=self):
        _WORKER_MANAGERS.manager = manager
        try:
          future.set_result(f())
        except BaseException as e:  
          future.set_exception(e)

      builder = lambda t, n: threading.Thread(target=t, name=n)
      thread = builder(run_inner, name)
      thread.setDaemon(True)

      thread.start()
      self._workers_count[name] += 1
      worker = ThreadWorker(thread=thread, future=future)
      self._active_workers[name].append(worker)

  def process_worker(self, name, command, env=None, **kwargs):
    """Adds process worker to the runtime.

    Args:
      name: Name of the worker's group.
      command: Command to execute in the worker.
      env: Environment variables to set for the worker.
      **kwargs: Other parameters to be passed to `subprocess.Popen`.
    """
    with self._mutex:
      process = subprocess.Popen(command, env=env or {}, **kwargs)  
      self._workers_count[name] += 1
      self._active_workers[name].append(process)

  def register_existing_process(self, name: str, pid: int):
    """Registers already started worker process.

    Args:
      name: Name of the workers' group.
      pid: Pid of the process to monitor.
    """
    with self._mutex:
      self._workers_count[name] += 1
      self._active_workers[name].append(psutil.Process(pid))

  def _stop_by_user(self):
    """Handles stopping of the runtime by a user."""
    if self._termination_notice_secs != 0:
      print(
          termcolor.colored(
              'User-requested termination. Asking workers to stop.', 'blue'))
      print(termcolor.colored('Press CTRL+C to terminate immediately.', 'blue'))
    signal.signal(signal.SIGINT, lambda sig, frame: self._kill())
    self._stop()

  def _kill_process_tree(self, pid):
    """Kills all child processes of the current process."""
    parent = psutil.Process(pid)
    for process in parent.children(recursive=True):
      try:
        process.send_signal(signal.SIGKILL)
      except psutil.NoSuchProcess:
        pass
    parent.send_signal(signal.SIGKILL)

  def _kill(self):
    """Kills all workers (and main thread/process if needed)."""
    print(termcolor.colored('\nKilling entire runtime.', 'blue'))
    kill_self = self._kill_main_thread
    for workers in self._active_workers.values():
      for worker in workers:
        if isinstance(worker, ThreadWorker):
          # Not possible to kill a thread without killing the process.
          kill_self = True
        else:
          self._kill_process_tree(worker.pid)
    if kill_self:
      self._kill_process_tree(os.getpid())

  def _stop_or_kill(self):
    """Stops all workers; kills them if they don't stop on time."""
    pending_secs = self._termination_notice_secs - self._stop_counter
    if pending_secs == 0:
      if self._termination_notice_secs > 0:
        still_running = [
            label for label in self._active_workers
            if self._active_workers[label]
        ]
        print(
            termcolor.colored(
                f'Worker groups that did not terminate in time: {still_running}',
                'red'))
      self._kill()
      return
    if pending_secs >= 0:
      print(
          termcolor.colored(f'Waiting for workers to stop for {pending_secs}s.',
                            'blue'),
          end='\r')
    self._stop_counter += 1
    for workers in self._active_workers.values():
      for worker in workers:
        if isinstance(worker, ThreadWorker):
          # Thread workers should use wait_for_stop.
          pass
        elif isinstance(worker, subprocess.Popen):
          worker.send_signal(signal.SIGTERM)
        else:
          # Notify all workers running under a proxy process.
          children = worker.children(recursive=True)
          worker_found = False
          for process in children:
            try:
              process_name = process.name()
              if process_name != 'bash' and 'envelope_' not in process_name:
                worker_found = True
                process.send_signal(signal.SIGTERM)
            except psutil.NoSuchProcess:
              pass
          if not worker_found:
            # No more workers running, so we can kill the proxy itself.
            try:
              worker.send_signal(signal.SIGKILL)
            except psutil.NoSuchProcess:
              pass

    if pending_secs >= 0:
      signal.alarm(1)

  def _stop(self):
    """Requests all workers to stop and schedule delayed termination."""
    self._stop_event.set()
    try:
      if self._termination_notice_secs > 0:
        self._alarm_enabled = True
        self._sigalrm_handler = register_signal_handler(
            signal.SIGALRM, lambda sig=None, frame=None: self._stop_or_kill())
    except ValueError:
      # This happens when we attempt to register a signal handler but not in the
      # main thread. Send a SIGTERM to redirect to the main thread.
      psutil.Process(os.getpid()).send_signal(signal.SIGTERM)
      return

    self._stop_or_kill()

  def _disable_alarm(self):
    if hasattr(self,'_alarm_enabled'):
      if self._alarm_enabled:
        self._alarm_enabled = False
        signal.alarm(0)
        remove_signal_handler(signal.SIGALRM, self._sigalrm_handler)
    else:
      self._sigalrm_handler = None
      self._alarm_enabled = False
      signal.alarm(0)
      remove_signal_handler(signal.SIGALRM, self._sigalrm_handler)

  def stop_and_wait(self):
    """Requests stopping all workers and wait for termination."""
    with self._mutex:
      self._stop()
    self.wait(raise_error=False)

  def join(self):
    self.wait()

  def wait(self,
           labels_to_wait_for: Optional[Sequence[Text]] = None,
           raise_error=True,
           return_on_first_completed=False):
    """Waits for workers to finish. Also stops the program upon worker failures.

    Args:
      labels_to_wait_for: If supplied, only wait for these groups' workers to
        finish. Wait for all workers otherwise.
      raise_error: Raise an exception upon any worker failure.
      return_on_first_completed: Whether to return upon the first completed (or
        failed) worker.

    Raises:
      RuntimeError: if any worker raises an exception.
    """
    active_workers = True
    while active_workers:
      with self._mutex:
        self._check_workers()
        active_workers = False
        if self._first_failure and raise_error:
          failure = self._first_failure
          self._first_failure = None
          raise failure
        for label in labels_to_wait_for or self._active_workers.keys():
          if self._active_workers[label]:
            active_workers = True
          if (return_on_first_completed and len(self._active_workers[label])
              < self._workers_count[label]):
            return
      time.sleep(0.1)

  def cleanup_after_test(self, test_case: absltest.TestCase):
    """Cleanups runtime after a test."""
    with self._mutex:
      self._check_workers()
      self._stop()
      self._disable_signals()
    self.wait(raise_error=False)
    with self._mutex:
      if self._first_failure:
        raise self._first_failure


  def check_for_thread_worker_exception(self):
    with self._mutex:
      for label in self._active_workers:
        for worker in self._active_workers[label]:
          if not worker.thread.is_alive():
            worker.thread.join()
            # This will raise the exception, if any.
            worker.future.result()

  def _check_workers(self):
    """Checks status of running workers, terminate runtime in case of errors.

    This REQUIRES holding self._mutex.
    """
    has_workers = False
    for label in self._active_workers:
      still_active = []
      for worker in self._active_workers[label]:
        active = True
        if isinstance(worker, ThreadWorker):
          if not worker.thread.is_alive():
            worker.thread.join()
            if not self._stop_counter:
              try:
                worker.future.result()
              except BaseException as e:  
                if not self._first_failure and not self._stop_counter:
                  self._first_failure = e
            active = False
        elif isinstance(worker, subprocess.Popen):
          try:
            res = worker.wait(0)
            active = False
            if res and not self._first_failure and not self._stop_counter:
              self._first_failure = RuntimeError('One of the workers failed.')
          except subprocess.TimeoutExpired:
            pass
        else:
          try:
            # We can't obtain return code of external process, so clean
            # termination is assumed.
            res = worker.wait(0)
            active = False
          except psutil.TimeoutExpired:
            pass
        if active:
          has_workers = True
          still_active.append(worker)
      self._active_workers[label] = still_active
    if has_workers and self._first_failure and not self._stop_counter:
      self._stop()
    elif not has_workers:
      self._disable_alarm()

  def __del__(self):
    self._disable_signals()



import launchpad as lp
from launchpad import address as lp_address
from launchpad.nodes.courier import courier_utils

from . import common


class CourierNode(lp.CourierNode):
  """Inherits `CourierNode` from launchpad and remove `worker_manager` from it."""
  def run(self) -> None:
    instance = self._construct_instance()  # pytype:disable=wrong-arg-types
    self._server = courier_utils.make_courier_server(
        instance,
        port=lp_address.get_port_from_address(self._address.resolve()),
        **self._courier_kwargs)
    if hasattr(instance, 'set_courier_server'):
      # Transfer the ownership of the server to the instance, so that the user
      # can decide when to start and stop the courier server.
      instance.set_courier_server(self._server)
      if hasattr(instance, 'run') and self._should_run:
        instance.run()
    else:
      # Start the server after instantiation and serve forever
      self._server.Start()
      if hasattr(instance, 'run') and self._should_run:
        # If a run() method is provided, stop the server at the end of run().
        instance.run()
      else:
        common.wait_for_stop()
      self._server.Stop()
      self._server.Join()



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Placeholders of network addresses to be evaluated at runtime."""

import abc
import os
import re
from typing import Optional
from absl import logging
import portpicker

_ADDRESS_NAME_VALID_PATTERN = re.compile('[a-z][a-z0-9]*')


class AbstractAddressBuilder(metaclass=abc.ABCMeta):
    """Base class for creating a platform-specific address at runtime."""

    @abc.abstractmethod
    def build(self) -> str:
        """Builds an address."""


class SimpleLocalAddressBuilder(AbstractAddressBuilder):
    """Creates a locahost:port address, with port decided by portpicker."""

    def __init__(self):
        # This automatically makes use of PORTSERVER_ADDRESS (usually set by test)
        self._address = 'localhost:{}'.format(portpicker.pick_unused_port())

    def build(self) -> str:
        return self._address


# 本机运行使用本机ip，即127.0.0.1
class LocalAddressBuilder(SimpleLocalAddressBuilder):
    def __init__(self):
        # This automatically makes use of PORTSERVER_ADDRESS (usually set by test)
        self._address = '127.0.0.1:{}'.format(portpicker.pick_unused_port())


class CustomAddress:
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.client_port = portpicker.pick_unused_port()

    def split(self, split_char):
        if split_char == ':':
            return [self.host, self.client_port]

    def __getitem__(self, item):
        if item == "host":
            return self.host
        elif item == "port":
            return self.port
        elif item == "client_port":
            return self.client_port
        else:
            return None
    def __str__(self):
        return "host:{},host port:{},client port:{}".format(self.host,self.port,self.client_port)

class HostAddressBuilder(SimpleLocalAddressBuilder):
    def __init__(self, host, port):
        # This automatically makes use of PORTSERVER_ADDRESS (usually set by test)
        # self._address = '{}:{}'.format(host, portpicker.pick_unused_port())
        self._address = CustomAddress(host,port)


class Address(object):
    """A network address to be evaluated.

    1. An unbound address is created using `address = Address()`.
    2. An address should be assigned to exactly one node e.g. using
       `address.assign(node)`.
    2. Upon launching, Launchpad will call bind() for each address to assign the
       address builder, which constructs the actual string format address at
       runtime. Launchpad has access to the addressed used in a program because
       each node maintains an `addresses` list of the nodes he knows about (it can
       be the address(es) the node own, or addresses that sub-nodes own.
    3. At runtime, use `address.resolve()` to finally resolve it.

    Using an unbound address will trigger an error.

    Attributes:
      name: Name of this address.
    """

    def __init__(self, name: Optional[str] = None):
        """Initializes an address object.

        Args:
          name: (Optional) Name of the address.
        """
        if name is not None and not _ADDRESS_NAME_VALID_PATTERN.fullmatch(name):
            raise ValueError(f'Wrong address name: {name} does not match '
                             f'{_ADDRESS_NAME_VALID_PATTERN.pattern}.')
        self.name = name
        self._address_builder = None  # type: AbstractAddressBuilder
        self._owning_node = None

    def bind(self, address_builder: AbstractAddressBuilder) -> None:
        """Sets a function that creates the platform-specific address at runtime."""
        # The address cannot be evaluated before we launch, because we might not
        # have all the necessary info for evaluation
        self._address_builder = address_builder

    def resolve(self) -> str:
        """Returns the address as a string."""
        if not self._address_builder:
            if self._owning_node is None:
                raise RuntimeError(
                    "The lp.Address hasn't been assigned to any node, "
                    'thus it cannot be resolved. Use '
                    '`address.assign(node)` to assign an address to a node')
            raise RuntimeError(
                f'The lp.Address associated to the node {self._owning_node} has not '
                'been bound to any address builder. Launchpad is responsible for '
                'doing that at launch time. If you are in tests, '
                'you can use:\n\n'
                'from launchpad.launch.test_multi_threading import address_builder as test_address_builder\n'
                '...\n'
                'test_address_builder.bind_addresses([node])')

        return self._address_builder.build()

    def assign(self, node) -> None:
        """Assigns the Address to the specified node (must be done exactly once)."""
        if self._owning_node is not None:
            if self._owning_node is node:
                logging.warning(
                    'You are binding an lp.Address twice on the same node '
                    "of type %s it's probably a mistake.", node)
                return
            raise ValueError('You are associating a node to this lp.Address which '
                             'is already assigned to another node. The previous node '
                             'and the new node are:\n'
                             f'{node}\n{self._owning_node}')
        self._owning_node = node
        node.addresses.append(self)

    def __getstate__(self):
        state = self.__dict__.copy()
        # Don't pickle `_owning_node`, as it's here only for launch-time checks.
        state['_owning_node'] = ('Restored from pickle node named: ' +
                                 repr(self._owning_node))
        return state


class HostAddress(Address):
    def to_host(self, host):
        if ":" in host:
            self.host = host.split(":")[0]
            self.port = host.split(":")[1]
        else:
            self.host = host
            self.port = "22"


def get_port_from_address(address: str) -> int:
    """Returns the port from a given address.

    Note that Launchpad uses a convention where named ports are passed as
    environment variables. For example, for the named port 'baz', the actual port
    value will be stored in LP_PORT_baz environment variable.

    Args:
      address: address with a named port or a host:port address.

    Returns:
      The port number as an integer.
    """
    port_name = address.split(':')[-1]
    if isinstance(port_name, int):
        return port_name
    elif port_name.isdigit():
        return int(port_name)
    else:
        return int(os.environ['LP_PORT_' + port_name])



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
import os
from tlaunch.lp_k8s.runner.script_runner import ScriptRunner
from tlaunch.lp_k8s.util import get_ipv4_by_hostname



class DeepSpeedRunner(ScriptRunner):
    def __init__(self, cmd, cmd_args, pwd):
        super().__init__(cmd, cmd_args, pwd)
        pass






# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Platform-specific configuration on the node."""

import enum
import threading

from typing import Any, Callable, Optional


class LaunchType(enum.Enum):
  """The different launch types supported by Launchpad."""
  LOCAL_MULTI_THREADING = 'local_mt'
  # Launch locally using multiple processes. Can display logs from different
  # nodes in separate windows. The behavior can be controlled using the
  # `terminal` argument in `launch.launch`.
  LOCAL_MULTI_PROCESSING = 'local_mp'
  # Launch using multiple processes, as a test.
  TEST_MULTI_PROCESSING = 'test_mp'
  # Launch as a test using multiple threads (same as LOCAL_MULTI_THREADING but
  # terminates the process instead of dropping into PDB).
  TEST_MULTI_THREADING = 'test_mt'
  SSH_MULTI_PROCESSING = 'ssh_mp'


class LaunchContext(object):
  """Stores platform-specific launch config of a node.

  This is created and set on the node only at launch time.
  """

  def __init__(self):
    self._launch_type = None
    self._launch_config = None
    self._program_stopper = None
    self._is_initialized = False

  @property
  def launch_type(self) -> LaunchType:
    self._check_inititialized()
    return self._launch_type

  @property
  def launch_config(self) -> Any:
    self._check_inititialized()
    return self._launch_config

  @property
  def program_stopper(self) -> Callable[[], None]:
    self._check_inititialized()
    return self._program_stopper

  def _check_inititialized(self):
    if not self._is_initialized:
      raise RuntimeError(
          'Launch context is not yet initialized. It should be initialized by '
          'calling initialize() at launch time.')

  def initialize(self, launch_type: LaunchType, launch_config: Any,
                 program_stopper: Optional[Callable[[], None]] = None):
    self._launch_config = launch_config
    self._launch_type = launch_type
    self._program_stopper = program_stopper
    self._is_initialized = True


_LAUNCH_CONTEXT = threading.local()


def get_context():
  context = getattr(_LAUNCH_CONTEXT, 'lp_context', None)
  assert context, ("Launchpad context was not instantiated. Do you try to "
                   "access it outside of the main node's thread?")
  return context


def set_context(context: LaunchContext):
  _LAUNCH_CONTEXT.lp_context = context



from launchpad.address import AbstractAddressBuilder


class KubernetesAddressBuilder(AbstractAddressBuilder):
  """ Creates a `{STATEFUL_SET_NAME}-{POD_ID}.{SERVICE_NAME}` address.

  The process runs in a container, so use a fixed port.
  """
  def __init__(self, lp_job_name: str, group_name: str, id: int) -> None:
    self._address = f'{lp_job_name}-{group_name}-{id}.{lp_job_name}:8001'

  def build(self) -> str:
    return self._address



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import reverb

class Server:
    def __init__(self,all_args):
        self.all_args = all_args
        self.server = self.launch_server(self.all_args)

    def launch_server(self, all_args):
        running_programs = ["server", "server_learner", "whole"]
        if not all_args.program_type in running_programs:
            return None
        # launch server
        port = int(all_args.server_address.split(':')[1])
        print('Launch server at port:{}'.format(port))
        if all_args.distributed_type == "sync":
            # data_rate_limiter = reverb.rate_limiters.Queue(all_args.actor_num)
            # weight_rate_limiter = reverb.rate_limiters.MinSize(1)
            # weight_max_times_sampled = all_args.actor_num
            # data_max_size = all_args.actor_num

            tables = [reverb.Table(
                name='sync_signal',
                sampler=reverb.selectors.Uniform(),
                remover=reverb.selectors.Fifo(),
                max_size=all_args.actor_num,
                rate_limiter=reverb.rate_limiters.Queue(all_args.actor_num),
                max_times_sampled=1),
                reverb.Table(
                    name='id',
                    sampler=reverb.selectors.Uniform(),
                    remover=reverb.selectors.Fifo(),
                    max_size=all_args.actor_num,
                    rate_limiter=reverb.rate_limiters.Queue(all_args.actor_num),
                    max_times_sampled=1)]

            for i in range(all_args.actor_num):
                tables.append(reverb.Table(  # Replay buffer storing weight.
                    name='weight_{}'.format(i),
                    sampler=reverb.selectors.Uniform(),
                    remover=reverb.selectors.Fifo(),
                    max_size=1,
                    rate_limiter=reverb.rate_limiters.Queue(1)))
                tables.append(reverb.Table(  # Replay buffer storing experience.
                    name='replay_buffer_{}'.format(i),
                    sampler=reverb.selectors.Uniform(),
                    remover=reverb.selectors.Fifo(),
                    max_size=1,
                    rate_limiter=reverb.rate_limiters.Queue(1)))
        else:
            # async type is reminded for future
            raise NotImplemented

        server = reverb.Server(
            tables=tables,
            port=port
        )
        if all_args.program_type == "server":
            server.wait()
            exit()
        else:
            return server

    def stop(self):
        if self.server:
            self.server.stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
# import time
from typing import Any, Callable, Optional, Sequence

from absl import logging

import portpicker

import reverb

from tlaunch.lp_ssh import address as lp_address

from tlaunch.lp_ssh.launch import worker_manager

# from tlaunch.lp_ssh import context
# from tlaunch.lp_ssh.nodes.python import node as python
from tlaunch.lp_ssh.nodes import ssh_node
from tmarl.transmit.rl_tables import rl_tables


# from tmarl import transmit
from tmarl.transmit import client
from tmarl.transmit import launch_server
from tlaunch.lp_ssh.nodes import base
from tlaunch.lp_ssh.address import get_port_from_address

PriorityTablesFactory = Callable[[], Sequence[reverb.Table]]
CheckpointerFactory = Callable[[], reverb.checkpointers.CheckpointerBase]

TRANSMIT_PORT_NAME = 'transmit'
import subprocess


class TransmitHandle(base.Handle):
    """Handle of the TransmitNode.

    When dereferenced a transmit-Client is returned. This client should primarily be used for insert operations on the
    actors.
    """

    def __init__(self, address: lp_address.Address):
        self._address = address

    def dereference(self):
        address = self._address.resolve()
        if isinstance(address,str):
            host = address.split(':')[0]
            client_port = address.split(':')[1]
        else:
            host = address.split(':')[0]
            host_port = address['port']
            client_port = get_port_from_address(address)
        print("data server address:",address)
        # print(host)
        # print(host_port)
        # 反向代理，但是程序结束后不会自动断开，需要实现自动断开的功能
        reverse_proxy = False
        if host == '127.0.0.1' or host == 'localhost':
            reverse_proxy=False

        if reverse_proxy:

            try:
                local_port = portpicker.pick_unused_port()
                server_address = 'localhost:{}'.format(local_port)
                exe_list = ['ssh','-p', host_port, '-NL', '{}:localhost:{}'.format(local_port, client_port), host, '&']
                # logging.info(exe_list)
                subprocess.Popen(exe_list)
                print("Use reverse proxy:{}".format(" ".join(exe_list)))

            except subprocess.CalledProcessError as e:
                print(e.output.decode())
                raise e  # If `tmux new-session` failed for some other reason.

            logging.info('Transmit client connecting to: %s', address)
            logging.info('Transmit client using local address: {}'.format(server_address))
        else:
            server_address = '{}:{}'.format(host,client_port)
            logging.info('Transmit client connecting to server: %s', server_address)

        return client.Client(server_address)


class TransmitNode(ssh_node.SSHNode):
    """Represents a Transmit Server in a Launchpad program."""

    def __init__(self,
                 priority_tables_fn: Optional[PriorityTablesFactory] = lambda: [],
                 checkpoint_ctor: Optional[CheckpointerFactory] = None):
        super().__init__(self.run)
        self._priority_tables_fn = priority_tables_fn
        self._checkpoint_ctor = checkpoint_ctor
        self._address = lp_address.HostAddress(TRANSMIT_PORT_NAME)

        self.allocate_address(self._address)

    def to_host(self, host):
        super(TransmitNode, self).to_host(host)
        self._address.to_host(host)
        return self

    def create_handle(self):
        return self._track_handle(TransmitHandle(self._address))

    def run(self):
        priority_tables = self._priority_tables_fn()

        # priority_tables.append(reverb.Table(name='info',
        #                      sampler=reverb.selectors.Uniform(),
        #                      remover=reverb.selectors.Fifo(),
        #                      max_size=1,
        #                      rate_limiter=reverb.rate_limiters.Queue(1),
        #                      ))
        priority_tables.append(reverb.Table(name='info',
                                            sampler=reverb.selectors.Lifo(),
                                            remover=reverb.selectors.Fifo(),
                                            max_size=1,
                                            rate_limiter=reverb.rate_limiters.MinSize(1)
                                            ))

        if self._checkpoint_ctor is None:
            checkpointer = None
        else:
            checkpointer = self._checkpoint_ctor()

        self._server = launch_server.LaunchServer(
            tables=priority_tables,
            port=lp_address.get_port_from_address(self._address.resolve()),
            checkpointer=checkpointer)
        worker_manager.wait_for_stop()

    @property
    def reverb_address(self) -> lp_address.Address:
        return self._address


class RLTransmitNode(TransmitNode):
    # RL训练的数据服务器建立
    def __init__(self, argv,actor_num,
                 checkpoint_ctor: Optional[CheckpointerFactory] = None):
        self.actor_num = actor_num
        self.argv = argv
        super().__init__(self.rl_tables, checkpoint_ctor)

    def rl_tables(self):
        from tmarl.configs.config import get_config
        parser = get_config()
        all_args = parser.parse_known_args(self.argv)[0]
        all_args.actor_num = self.actor_num
        return rl_tables(all_args=all_args)




# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing all Launchpad flags."""

# from absl import flags
# from launchpad import context
# FLAGS = flags.FLAGS

from absl import flags
from . import context
FLAGS = flags.FLAGS

try:
    LP_TERMINATION_NOTICE_SECS = flags.DEFINE_integer(
        'lp_termination_notice_secs', 10,
        'Send termination notice to all nodes that many seconds before hard '
        'termination. Set to 0 to trigger hard termination righ away (skip '
        'termination notice), set to negative value to disable hard termination.')
except:
    pass







# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Nodes that run user-defined Python code.

PyNode runs a user-defined function. PyClassNode constructs a Python object and
calls its run() method (if provided).

"""

import functools
import itertools
from typing import Any, Callable, Generic, TypeVar

from absl import logging

from launchpad import lazy_loader

from launchpad.nodes import dereference
from tlaunch.lp_ssh.nodes.python import addressing

from launchpad.nodes.python import local_multi_processing
import tree

from tlaunch.lp_ssh.nodes import base
from tlaunch.lp_ssh import context

T = TypeVar('T')
HandleType = TypeVar('HandleType', bound=base.Handle)
ReturnType = TypeVar('ReturnType')
WorkerType = TypeVar('WorkerType')


class _DummyHandle(base.Handle[Any]):

  def dereference(self) -> None:
    raise NotImplementedError('_DummyHandle cannot be dereferenced.')


class PyNode(base.Node[HandleType], Generic[HandleType, ReturnType]):
  """Runs a user-defined Python function."""

  # Used only for no-serialization launch
  NEXT_PY_NODE_ID = itertools.count()

  def __init__(self, function: Callable[..., ReturnType], *args, **kwargs):
    super().__init__()
    self._func_args = args
    self._func_kwargs = kwargs
    self._function = function
    self._partial_function = self._construct_function
    self.py_node_id = next(PyNode.NEXT_PY_NODE_ID)

    # Find input handles and put them in self._input_handles.
    tree.map_structure(
        functools.partial(base.extract_handles, handles=self._input_handles),
        (self._func_args, self._func_kwargs))

  def _construct_function(self):
    context.set_context(self._launch_context)  
    args, kwargs = tree.map_structure(dereference.maybe_dereference,
                                      (self._func_args, self._func_kwargs))
    return functools.partial(self._function, *args, **kwargs)()

  def create_handle(self) -> HandleType:
    """Doesn't expose an interface for others to interact with it."""
    return _DummyHandle()

  @property
  def function(self) -> Callable[..., ReturnType]:
    return self._partial_function

  @staticmethod
  def to_executables(nodes, label, launch_context):
    """Creates Executables."""
    if (launch_context.launch_type in [
        context.LaunchType.LOCAL_MULTI_THREADING,
        context.LaunchType.TEST_MULTI_THREADING
    ]):
      return [node.function for node in nodes]
    elif (launch_context.launch_type is
          context.LaunchType.LOCAL_MULTI_PROCESSING):
      return local_multi_processing.to_multiprocessing_executables(
          nodes, label, launch_context.launch_config, pdb_post_mortem=True)
    elif (
        launch_context.launch_type is context.LaunchType.TEST_MULTI_PROCESSING):
      return local_multi_processing.to_multiprocessing_executables(
          nodes, label, launch_context.launch_config, pdb_post_mortem=False)
    raise NotImplementedError('Unsupported launch type: {}'.format(
        launch_context.launch_type))

  def bind_addresses(self, **kwargs):
    if self._launch_context.launch_type in [
        context.LaunchType.LOCAL_MULTI_THREADING,
        context.LaunchType.LOCAL_MULTI_PROCESSING,
        context.LaunchType.TEST_MULTI_PROCESSING,
        context.LaunchType.TEST_MULTI_THREADING
    ]:
      addressing.bind_addresses_local(self.addresses)
    elif self._launch_context.launch_type in [context.LaunchType.SSH_MULTI_PROCESSING,]:
      addressing.bind_addresses_host(self.addresses)
    else:
      raise NotImplementedError('Unsupported launch type: {}'.format(
          self._launch_context.launch_type))

  @classmethod
  def default_launch_config(cls, launch_type: context.LaunchType):
    if launch_type in [
        context.LaunchType.LOCAL_MULTI_THREADING,
        context.LaunchType.TEST_MULTI_THREADING
    ]:
      return None
    return super().default_launch_config(launch_type)


class PyClassNode(PyNode[HandleType, type(None)],
                  Generic[HandleType, WorkerType]):
  """Instantiates a Python object and runs its run() method (if provided).

  If disable_run() is called before launch, instance.run() method won't be
  called. This is useful in TAP-based integration tests, where users might need
  to step each worker synchronously.
  """

  def __init__(self, constructor: Callable[..., WorkerType], *args, **kwargs):
    """Initializes a new instance of the `PyClassNode` class.

    Args:
      constructor: A function that when called returns a Python object with a
        run method.
      *args: Arguments passed to the constructor.
      **kwargs: Key word arguments passed to the constructor.
    """
    super().__init__(self.run)
    self._constructor = constructor
    self._args = args
    self._kwargs = kwargs
    self._should_run = True
    self._collect_input_handles()

  def _collect_input_handles(self):
    self._input_handles.clear()
    try:
      # Find input handles and put them in self._input_handles.
      tree.map_structure(
          functools.partial(base.extract_handles, handles=self._input_handles),
          (self._args, self._kwargs))
    except TypeError as e:
      raise ValueError(
          f'Failed to construct the {self.__class__.__name__} with\n'
          f'- constructor: {self._constructor}\n'
          f'- args: {self._args}\n- kwargs: {self._kwargs}') from e

  def _construct_instance(self) -> WorkerType:
    args, kwargs = tree.map_structure(dereference.maybe_dereference,
                                      (self._args, self._kwargs))
    return self._constructor(*args, **kwargs)

  def disable_run(self) -> None:
    """Prevents the node from calling `run` on the Python object.

    Note that the Python object is still constructed even if `disable_run` has
    been called.
    """
    self._should_run = False

  def enable_run(self) -> None:
    """Ensures `run` is called on the Python object.

    This is the default state and callers don't need to call `enable_run` unless
    `disable_run` has been called.
    """
    self._should_run = True

  def run(self) -> None:
    """Constructs Python object and (maybe) calls its `run` method.

    The `run` method is not called if `disable_run` has ben called previously or
    if the constructed Python object does not have a `run` method.
    """
    instance = self._construct_instance()
    if hasattr(instance, 'run') and self._should_run:
      instance.run()
    else:
      logging.warning(
          'run() not defined on the instance (or disable_run() was called.).'
          'Exiting...')



# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A Launchpad Node.

Node represents a service. It may return a handle for others to interact with
it.
"""

import abc
import functools
from typing import Any, Generic, List, Optional, Sequence, Set, TypeVar

from launchpad import address as lp_address
from tlaunch.lp_ssh import context as lp_context
from launchpad.nodes import dereference
from tlaunch.lp_ssh.program_stopper import program_stopper

ClientType = TypeVar('ClientType')
HandleType = TypeVar('HandleType', bound='Handle')


class Handle(dereference.Dereferenceable[ClientType], Generic[ClientType]):
  """Represents an interface of the service (the node).

  Call `.dereference()` to get the actual client object of this service (to be
  implemented in subclasses).
  """

  def connect(self, node: 'Node[Handle[ClientType]]', label: str) -> None:
    """Called to let this handle know about it's connecting to a node.

    This is supposed to be called:

      1. Before creating any executables
      2. Before any address binding happens

    The motivation is we want to give the handle a chance to configure itself
    for the node, before it's turned into executables and addresses are
    finalized.

    Args:
      node: The node that the handle connects to.
      label: Label of the node.
    """
    pass

  def transform(self, executables: Sequence[Any]) -> Sequence[Any]:
    """Transforms the executables that make use of this handle."""
    return executables


class Node(Generic[HandleType], metaclass=abc.ABCMeta):
  """Represents a service, and may return a Handle for interaction with it."""

  def __init__(self) -> None:
    # This is a low-level API to allow Node/Handle to access launch config
    # during run time. It's only available after launch, and it's set by the
    # launcher.
    self._launch_context = lp_context.LaunchContext()
    # Handles used by this node (to interact with other nodes)
    self._input_handles = []  # type: List[Handle[Any]]
    # Handles created by this node
    # Note: `type: List[HandleType]` is not supported yet.
    self._created_handles = []  # type: List[Handle]
    # Addresses owned by this node (i.e., addresses used to run servers).
    self.addresses = []  # type: List[lp_address.Address]

  def _initialize_context(self, launch_type: lp_context.LaunchType,
                          launch_config: Any):
    self._launch_context.initialize(
        launch_type, launch_config,
        program_stopper.make_program_stopper(launch_type))

  def _track_handle(self, handle: HandleType) -> HandleType:
    """Keeps track of created handles.

    MUST be called in create_handle().

    This is called so that the node knows about the handle it creates. The
    reason we don't automate this is because we'll lose return annotation if
    we wrap create_handle() using a base class method (i.e., the base class
    wrapper method doesn't know about the subclass return type).

    Args:
      handle: The handle (MUST be created by this node) to track.

    Returns:
      The same handle that was passed in, for the nicer syntax on call site.
    """
    self._created_handles.append(handle)
    return handle

  @abc.abstractmethod
  def create_handle(self) -> HandleType:
    """Creates a handle to interact with this node.

    MUST call _track_handle() after creating a handle.
    """
    raise NotImplementedError()

  @abc.abstractstaticmethod
  def to_executables(nodes, label, context):
    """Creates executables for a specific launch type."""
    raise NotImplementedError()


  def bind_addresses(self, **kwargs) -> None:
    """Binds addresses of the node."""
    del kwargs  # Unused.

  @property
  def input_handles(self) -> List[Handle[Any]]:
    return list(self._input_handles)

  def allocate_address(self, address: lp_address.Address) -> None:
    """Low-level API to add an address to listen to.

    This is a low level API and users shouldn't need to use it most of the time.

    Args:
      address: Address to listen to (i.e., to create a server).
    """
    self.addresses.append(address)

  @classmethod
  def default_launch_config(cls, launch_type: lp_context.LaunchType):
    """Defines the default launch config of this node type.

    This is optional. The returned config is conditional on the launch type.

    Args:
      launch_type: Return the launch_config for this launch_type.
    """
    raise NotImplementedError(
        f'Launch config has to be explicitly specified for {cls.__name__}')


def extract_handles(
    obj: Any,
    handles: List[Handle],
    visited: Optional[Set[int]] = None,
) -> None:
  """Extract the handles of `obj` to and add them into `handles`."""
  visited = visited or set()
  # Transitive input_handles from Deferred objects are included.
  if isinstance(obj, dereference.Deferred):
    if id(obj) not in visited:
      visited.add(id(obj))
      obj._apply_to_args(  
          functools.partial(extract_handles, handles=handles, visited=visited))
  elif isinstance(obj, Handle):
    handles.append(obj)



from typing import Any, Mapping, Optional

from tlaunch.lp_ssh import context
from launchpad import program as lp_program
from launchpad.launch.run_locally import run_locally


def launch(program: lp_program.Program,
           local_resources: Optional[Mapping[str, Any]] = None,
           terminal: str = 'gnome-terminal'):
  """Launches a program using multiple processes."""
  # Set up the launch context (launch type & launch config) for all nodes
  local_resources = local_resources or {}
  for label, nodes in program.groups.items():
    launch_config = local_resources.get(label, None)
    for node in nodes:
      node._initialize_context(
        context.LaunchType.LOCAL_MULTI_PROCESSING,
        launch_config=launch_config)

  # Notify the input handles
  for label, nodes in program.groups.items():
    for node in nodes:
      for handle in node._input_handles:
        handle.connect(node, label)

  # Bind addresses
  for node in program.get_all_nodes():
    node.bind_addresses()

  commands = []
  for label, nodes in program.groups.items():
    # to_executables() is a static method, so we can call it from any of the
    # nodes in this group.

    # pytype: disable=wrong-arg-count
    commands.extend(nodes[0].to_executables(nodes, label,
                                            nodes[0]._launch_context))

    # pytype: enable=wrong-arg-count
  return run_locally.run_commands_locally(commands, terminal)



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A Launchpad program."""
import argparse
import contextlib
import itertools

from typing import Any, Dict, List, Optional

from .nodes import base

HandleType = Any

class Program(object):
  """A Launchpad program."""

  def __init__(self, name: str,all_args:argparse.Namespace):
    self._name = name
    self._groups = {}  # type: Dict[str, List[base.Node]]
    # Group to add nodes to. Used by group()
    self._current_group = None  # type: str
    self.all_args = all_args

  def add_node(self,
               node: base.Node,
               label: Optional[str] = None) -> HandleType:
    """Adds node to the program."""

    if self._current_group:
      if label and label != self._current_group:
        raise ValueError('The given label does not match the current group: '
                         f'{label} vs {self._current_group}.')
      label = self._current_group
    else:
      if not label:
        raise ValueError('Label should not be empty.')
    if label not in self._groups:
      self._groups[label] = [node]
    else:
      self._groups[label].append(node)
    return node.create_handle()

  @contextlib.contextmanager
  def group(self, label: str):
    """Creates a group for a collection of homogeneous nodes."""
    if not label:
      raise ValueError('Label should not be empty.')
    if self._current_group:
      raise ValueError('group() cannot be nested.')
    try:
      self._current_group = label
      yield
    finally:
      # Try/finally is to make sure that the current_group is correctly
      # reset even if an exception occurs.
      self._current_group = None


  def get_all_nodes(self) -> List[base.Node]:
    return list(itertools.chain(*self._groups.values()))

  @property
  def name(self) -> str:
    return self._name

  @property
  def groups(self) -> Dict[str, List[base.Node]]:
    return self._groups



def make_program(*nodes: base.Node, name: str = 'launchpad'):
  """A shortcut to create a program from a list of nodes.

  This simplifies the syntax. For example you can do a one-liner launch:

      lp.launch(lp.make_program(lp.PyNode(lambda: ...)))

  Args:
    *nodes: Nodes to run.
    name: An optional name of the program.

  Returns:
    A lp.Program object
  """
  program = Program(name)
  for node in nodes:
    program.add_node(node)
  return program



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import os
from setuptools import setup, find_packages
import setuptools

def get_version() -> str:
    # https://packaging.python.org/guides/single-sourcing-package-version/
    init = open(os.path.join("tlaunch", "__init__.py"), "r").read().split()
    return init[init.index("__version__") + 2][1:-1]


setup(
    name="tlaunch",  # Replace with your own username
    version=get_version(),
    description="tlaunch",
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    author="tlaunch",
    author_email="tmarl_contact@tartrl.cn",
    packages=setuptools.find_packages(),
    install_requires=[
        "protobuf==3.20.0",
        "dm-reverb==0.7.0",
        "dill",
        "dm-launchpad==0.5.0",
        "tensorflow_gpu==2.8.0",
        "cloudpickle",
        "kubernetes",
        "minio"
    ],
    entry_points={'console_scripts': ['tlaunchrun = tlaunch.lp_k8s.runner.run:main']},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    keywords="multi-agent reinforcement learning algorithms pytorch",
    python_requires='>=3.6',
)



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for command line flags."""

import sys
from typing import Any, Dict, List

from absl import flags

from tlaunch.lp_ssh.flags import flags as lp_flags


FLAGS = flags.FLAGS


def _flag_to_placeholder(key):
  """Maps flag_type to a value so that entry script can add flag definition."""
  flag_type = FLAGS[key].flag_type()
  flag_type_to_placeholder = {
      FLAGS['lp_dummy_str'].flag_type(): FLAGS.lp_dummy_str,
      FLAGS['lp_dummy_float'].flag_type(): FLAGS.lp_dummy_float,
      FLAGS['lp_dummy_int'].flag_type(): FLAGS.lp_dummy_int,
      FLAGS['lp_dummy_bool'].flag_type(): FLAGS.lp_dummy_bool,
      FLAGS['lp_dummy_list'].flag_type(): FLAGS.lp_dummy_list,
      FLAGS['lp_dummy_enum'].flag_type(): FLAGS.lp_dummy_enum,
      FLAGS['lp_dummy_config'].flag_type(): {},
  }

  if flag_type not in flag_type_to_placeholder:
    raise NotImplementedError(f'Flag type {flag_type} is not yet supported.')
  placeholder = flag_type_to_placeholder[flag_type]
  return placeholder


def _maybe_parse_jax_flags():
  if 'jax' in sys.modules and not sys.modules['jax'].config.use_absl:
    # JAX doesn't add absl flags by default. Making them visible in flags.FLAGS
    # allows us to figure out the correct flag types.
    sys.modules['jax'].config.config_with_absl()


def get_flags_to_populate(args: List[Any]) -> Dict[str, Any]:
  """Finds flags that require definitions for the default entry script."""
  _maybe_parse_jax_flags()
  flags_to_populate = {}
  for arg in args:
    if not isinstance(arg, tuple):
      continue
    key, _ = arg
    # Additional C++ flags, possibly linked only in the worker processes.
    if key not in FLAGS or FLAGS[key].flag_type() == '[C++]':
      continue
    flags_to_populate[key] = _flag_to_placeholder(key)
  return flags_to_populate



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

from typing import Callable, Optional, Sequence
import launchpad as lp
import reverb

from ...node.reverb import ReverbNode, ReverbHandle
from ...node import common

PriorityTablesFactory = Callable[[], Sequence[reverb.Table]]
CheckpointerFactory = Callable[[], reverb.checkpointers.CheckpointerBase]

TRANSMIT_PORT_NAME = 'transmit'


class TransmitNode(ReverbNode):
    """Represents a Transmit Server in a Launchpad program."""

    def __init__(self, Client, Server, *args, **kwargs):
        self.Client = Client
        self.Server = Server
        super().__init__(*args, **kwargs)

    def run(self):
        priority_tables = self._priority_tables_fn()
        priority_tables.append(reverb.Table(name='info',
                                            sampler=reverb.selectors.Uniform(),
                                            remover=reverb.selectors.Fifo(),
                                            max_size=1,
                                            rate_limiter=reverb.rate_limiters.Queue(1),
                                            ))

        if self._checkpoint_ctor is None:
            checkpointer = None
        else:
            checkpointer = self._checkpoint_ctor()

        self._server = self.Server(
            tables=priority_tables,
            port=lp.get_port_from_address(
                self._address.resolve()),
            checkpointer=checkpointer)
        common.wait_for_stop()

    def create_handle(self):
        return self._track_handle(ReverbHandle(self._address, self.Client))


from typing import NamedTuple, Optional, Tuple, Union


class ResourceRange(NamedTuple):
  requests: Optional[str]
  limits: Optional[str]


R = Union[ResourceRange, Tuple[str, Optional[str]], str, int, None]


class Resource:
  def __init__(self,
               cpu: R = None,
               memory: R = None,
               nvidia_gpu: R = None,
               nvidia_gpu_memory: R = None,
               nvidia_gpu_cores: R = None) -> None:
    self.resources = {
        'cpu': cpu,
        'memory': memory,
        'nvidia.com/gpu': nvidia_gpu,
        'nvidia.com/gpumem': nvidia_gpu_memory,
        'nvidia.com/gpucores': nvidia_gpu_cores
    }
    self.requests = {
        k: v if isinstance(v, str) else v[0] if isinstance(v, tuple) or isinstance(v, list) else v
        for k, v in self.resources.items() if v is not None
    }
    self.limits = {
        k: v if isinstance(v, str) else v[1] if isinstance(v, tuple) or isinstance(v, list) else v
        for k, v in self.resources.items() if v is not None
    }



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
import launchpad as lp
import os
import sys
import uuid
import time
from argparse import REMAINDER, ArgumentParser
from tlaunch.lp_k8s.util import generate_tlaunchjob_name, get_namespace

def get_args_parser() -> ArgumentParser:
    """Helper function parsing the command line options."""

    parser = ArgumentParser(description="TLaunch Distributed Training Launcher")
    parser.add_argument("--lpjob_name", default=generate_tlaunchjob_name(), type=str, help="The name of this job")
    parser.add_argument("--namespace", default=get_namespace(), type=str, help="The namespace of this job")
    parser.add_argument("--image", default='docker.4pd.io/tlaunch/reverb:base', type=str, help="The docker image \
            of trainer")
    parser.add_argument("--image_pull_policy", choices=['Always','IfNotPresent','Never'], default="Always", type=str,
            help="The pull policy of image")
    parser.add_argument("--gpu", default=2, type=int, help="The number of gpu per trainer")
    parser.add_argument("--gpu_memory", default=3000, type=int, help="The memory size of each gpu")
    parser.add_argument("--gpu_cores", default=30, type=int, help="The maximum available cores of each gpu")
    parser.add_argument("--set_save_path", action='store_true', default=False, help="Whether to automatically set a \
            path to save the model, if use this option, TLaunch will generate a folder in TData/model and add a argument \
            --save_path for all trainer pods, which is pointing to this folder")
    parser.add_argument("--set_checkpoint_path", action='store_true', default=False, help="Whether to automatically set a \
            path to save the checkpoint file, if use this option, TLaunch will generate a file named last.tar.gz in \
            TData/model and add a argument --checkpoint_path for all trainer pods, which is pointing to this file. \
            This parameter is usually used to store checkpoint and elastic deployment")
    subparsers = parser.add_subparsers()

    # basic launcher
    parser_basic = subparsers.add_parser('basic')
    parser_basic.add_argument("--num_trainer", default=1, type=int, help="The number of trainer")
    parser_basic.add_argument("--set_share_file", action='store_true', default=False, help="Whether to automatically set \
            a cache file for transferring distributed shared information, if use this option, TLaunch will generate \
                pointing to this file")
    parser_basic.set_defaults(func=k8s_launch)
    parser_basic.add_argument("training_script_args", nargs=REMAINDER)

    # torchrun launcher
    parser_torchrun = subparsers.add_parser('torchrun')
    parser_torchrun.add_argument("--nnodes", default='1', type=str)
    parser_torchrun.add_argument("--nproc_per_node", default=1, type=int)
    parser_torchrun.add_argument("--rdzv_id", default=1, type=int)
    parser_torchrun.add_argument("--rdzv_backend", default="c10d", type=str)
    parser_torchrun.set_defaults(func=k8s_torchrun_launch)
    parser_torchrun.add_argument("training_script_args", nargs=REMAINDER)

    # deepspeed launcher
    parser_deepspeed = subparsers.add_parser('deepspeed')
    parser_deepspeed.add_argument("--num_nodes", default='1', type=str)
    parser_deepspeed.add_argument("--num_gpus", default='1', type=int)
    parser_deepspeed.set_defaults(func=k8s_deepspeed_launch)
    parser_deepspeed.add_argument("training_script_args", nargs=REMAINDER)
    return parser


def parse_args(args):
    parser = get_args_parser()
    return parser.parse_args(args)

def set_base_args(args, cmd_args):
    from tlaunch import lp_k8s
    save_path = os.path.join(lp_k8s.config.DEFAULT_MNT_PATH, 'models', args.lpjob_name)
    checkpoint_path = os.path.join(save_path, 'last.tar.gz')
    if args.set_save_path:
        cmd_args.extend(['--save_path', save_path])
        if os.path.exists(save_path) is not True:
            os.mkdir(save_path)
    if args.set_checkpoint_path:
        if os.path.exists(save_path) is not True:
            os.mkdir(save_path)
        cmd_args.extend(['--checkpoint_path', checkpoint_path])

def k8s_deepspeed_launch(args):
    from tlaunch.lp_k8s.runner.deepspeed_runner import DeepSpeedRunner
    from tlaunch import lp_k8s
    from tlaunch.lp_k8s import Config, Container, Resource

    cmd = 'deepspeed'
    cmd_args = [f'--num_nodes={args.nnodes}',
                f'--num_gpus={args.gpu}'] + args.training_script_args
    pwd = os.getcwd()

    set_base_args(args, cmd_args)
    program = lp.Program(args.lpjob_name)
    configs = {}
    namespace = args.namespace
    node_cmd_args = cmd_args
    node = lp_k8s.CourierNode(DeepSpeedRunner,
                              cmd=cmd,
                              cmd_args=node_cmd_args,
                              pwd=pwd)
    program.add_node(node, label='trainer')
    configs['trainer'] = Config(namespace=namespace,
                                container=Container(namespace=namespace,
                                                    resources=Resource(nvidia_gpu=args.gpu,
                                                                       nvidia_gpu_memory=args.gpu_memory,
                                                                       nvidia_gpu_cores=args.gpu_cores),
                                                    image=args.image,
                                                    image_pull_policy=args.image_pull_policy))

    lp_k8s.launch(program, namespace=namespace, group_config=configs)


def k8s_torchrun_launch(args):
    from tlaunch.lp_k8s.runner.torch_runner import TorchRunner
    from tlaunch import lp_k8s
    from tlaunch.lp_k8s import Config, Container, Resource
    from tlaunch.lp_k8s.util import get_hostname_by_lpjob

    cmd = 'torchrun'
    cmd_args = ['--nnodes={}'.format(args.nnodes),
                '--nproc_per_node={}'.format(args.nproc_per_node),
                '--rdzv_id={}'.format(args.rdzv_id),
                '--rdzv_backend={}'.format(args.rdzv_backend)] + args.training_script_args
    pwd = os.getcwd()

    set_base_args(args, cmd_args)
    rdzv_endpoint = get_hostname_by_lpjob(args.lpjob_name, 'trainer', 0)

    program = lp.Program(args.lpjob_name)
    configs = {}
    namespace = args.namespace
    num_trainer = int(args.nnodes.split(':')[-1])
    for rank in range(num_trainer):
        node_cmd_args = cmd_args
        node = lp_k8s.CourierNode(TorchRunner,
                                  rdzv_endpoint=rdzv_endpoint,
                                  cmd=cmd,
                                  cmd_args=node_cmd_args,
                                  pwd=pwd)
        program.add_node(node, label='trainer')
    configs['trainer'] = Config(namespace=namespace,
                                container=Container(namespace=namespace,
                                                    resources=Resource(nvidia_gpu=args.gpu,
                                                                       nvidia_gpu_memory=args.gpu_memory,
                                                                       nvidia_gpu_cores=args.gpu_cores),
                                                    image=args.image,
                                                    image_pull_policy=args.image_pull_policy,
                                                    ports=[('torch', 8002)]))

    lp_k8s.launch(program, namespace=namespace, group_config=configs)


def k8s_launch(args):
    from tlaunch.lp_k8s.runner.script_runner import ScriptRunner
    from tlaunch import lp_k8s
    from tlaunch.lp_k8s import Config, Container, Resource

    cmd = os.getenv("PYTHON_EXEC", sys.executable)
    cmd_args = ["-u"] + args.training_script_args + \
               ['--world_size', args.num_trainer]
    pwd = os.getcwd()
    set_base_args(args, cmd_args)
    if args.set_share_file:
        share_file = os.path.join(lp_k8s.config.DEFAULT_MNT_PATH, 'cache', args.lpjob_name, str(uuid.uuid4()))
        cmd_args.extend(['--share_file', share_file])

    program = lp.Program(args.lpjob_name)
    configs = {}
    namespace = args.namespace
    for rank in range(args.num_trainer):
        node_cmd_args = cmd_args.copy()
        node_cmd_args.extend(['--rank', rank])
        node = lp_k8s.CourierNode(ScriptRunner,
                                  cmd=cmd,
                                  cmd_args=node_cmd_args,
                                  pwd=pwd)
        program.add_node(node, label='trainer')
        configs['trainer'] = Config(namespace=namespace,
                                    container=Container(namespace=namespace,
                                                        resources=Resource(nvidia_gpu=args.gpu,
                                                                           nvidia_gpu_memory=args.gpu_memory,
                                                                           nvidia_gpu_cores=args.gpu_cores),
                                                        image=args.image,
                                                        image_pull_policy=args.image_pull_policy))

    lp_k8s.launch(program, namespace=namespace, group_config=configs)


def main():
    args = parse_args(sys.argv[1:])
    args.func(args)



from .launch import launch
from .stop import stop
from .resource import Resource, ResourceRange
from .config import Config, Container, Volumer
from .node.reverb import ReverbNode
from .node.courier import CourierNode
from .source import Source, GitSource
from .volume_mount import VolumeMount
from .persistent_volume_claim_volume_source import PersistentVolumeClaimVolumeSource



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to run PyNodes as multiple processes."""

import atexit
import json
import os
import shutil
import sys
import tempfile
from typing import Any, List, Mapping, Optional, Union, Sequence, Tuple

from absl import flags
from absl import logging

import cloudpickle
import dataclasses

from tlaunch.lp_ssh.flags import flags as lp_flags
from launchpad.launch.local_multi_processing import commands as mp_commands
from launchpad.nodes.python import flags_utils

_INTERPRETER = sys.executable

StrOrFloat = Union[str, float]


def _to_cmd_arg(key: str, value: Any) -> Union[Tuple[str], Tuple[str, str]]:
  """Converts key value pair to "--key=value"."""
  if isinstance(value, bool):
    return ('--{}'.format(key) if value else '--no{}'.format(key),)
  return ('--{}'.format(key), str(value))




@dataclasses.dataclass
class PythonProcess:
  """Local multiprocessing launch configuration for a PyNode.

  Attributes:
    args: Arguments to pass to the user script.
    env: Additional environment variables to set.
  """

  args: Mapping[str, StrOrFloat] = dataclasses.field(default_factory=dict)
  env: Mapping[str, StrOrFloat] = dataclasses.field(default_factory=dict)

  _absolute_interpreter_path: str = ''

  def _get_absolute_interpreter_path(self):
    """Resolve self.interpreter to an absolute path."""
    return _INTERPRETER

  @property
  def absolute_interpreter_path(self) -> str:
    """Returns the absolute path to the interpreter binary."""
    if not self._absolute_interpreter_path:
      self._absolute_interpreter_path = self._get_absolute_interpreter_path()
    return self._absolute_interpreter_path


_DATA_FILE_NAME = 'job.pkl'


def to_multiprocessing_executables(
    nodes: Sequence[Any], label: str, launch_config: PythonProcess,
    pdb_post_mortem: bool) -> List[mp_commands.Command]:
  """Returns a list of `Command`s objects for the given `PyNode`s."""
  launch_config = launch_config or PythonProcess()
  if not isinstance(launch_config, PythonProcess):
    raise ValueError(
        'Launch config for {} must be a PythonProcess.'.format(label))


  entry_script_path = os.path.join(os.path.dirname(__file__),
                                   'process_entry.py')

  tmp_dir = tempfile.mkdtemp()
  atexit.register(shutil.rmtree, tmp_dir, ignore_errors=True)
  data_file_path = os.path.join(tmp_dir, _DATA_FILE_NAME)
  with open(data_file_path, 'wb') as f:
    cloudpickle.dump([node.function for node in nodes], f)


  commands = []
  for task_id, _ in enumerate(nodes):
    command_as_list = [
        launch_config.absolute_interpreter_path, entry_script_path
    ]


    # Arguments to pass to the script
    for key, value in launch_config.args.items():
      command_as_list.extend(_to_cmd_arg(key, value))

    # Find flags and pre-populate their definitions, as these definitions are
    # not yet ready in the entry script.
    flags_to_populate = flags_utils.get_flags_to_populate(
        list(launch_config.args.items()))
    if flags_to_populate:
      command_as_list.extend(
          _to_cmd_arg('flags_to_populate', json.dumps(flags_to_populate)))

    command_as_list.extend([
        '--data_file', data_file_path,
        '--lp_task_id', str(task_id),
    ])
    command = mp_commands.Command(command_as_list, launch_config.env,
                                  label + '/' + str(task_id))
    commands.append(command)

  return commands



import logging
from launchpad import context, Program
from kubernetes import client, config
from tlaunch.lp_k8s.client import validate_name

K8S_API_GROUP = "tartrl.cn"
K8S_API_VERSION =  "v1alpha1"
K8S_API_PLURAL = "lpjobs"


def stop(program_name: str = None, namespace: str = "default"):
    config.load_incluster_config()
    kube_api = client.CustomObjectsApi()
    if program_name is None:
        launch_config = context.get_context().launch_config
        namespace = launch_config["kube_namespace"]
        program_name = launch_config["lpjob_name"]
    kube_api.delete_namespaced_custom_object(
        K8S_API_GROUP,
        K8S_API_VERSION,
        namespace,
        K8S_API_PLURAL,
        validate_name(program_name))
    logging.info(f"An lpjob named '{program_name}' has been deleted in namespace '{namespace}'")
